#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Packing commands for odahuflow cli
"""
import logging
import time
from http.client import HTTPException

import click
from requests import RequestException
from odahuflow.cli.utils.client import pass_obj
from odahuflow.cli.utils.error_handler import check_id_or_file_params_present, TIMEOUT_ERROR_MESSAGE
from odahuflow.cli.utils.logs import print_logs
from odahuflow.cli.utils.output import format_output, DEFAULT_OUTPUT_FORMAT, validate_output_format
from odahuflow.sdk import config
from odahuflow.sdk.clients.api import WrongHttpStatusCode, APIConnectionException
from odahuflow.sdk.clients.api_aggregated import parse_resources_file_with_one_item
from odahuflow.sdk.clients.packaging import ModelPackaging, ModelPackagingClient, SUCCEEDED_STATE, FAILED_STATE

DEFAULT_WAIT_TIMEOUT = 3
# 1 hour
DEFAULT_PACKAGING_TIMEOUT = 60 * 60
LOG_READ_TIMEOUT_SECONDS = 60

LOGGER = logging.getLogger(__name__)


@click.group()
@click.option('--url', help='API server host', default=config.API_URL)
@click.option('--token', help='API server jwt token', default=config.API_TOKEN)
@click.pass_context
def packaging(ctx: click.core.Context, url: str, token: str):
    """
    Allow you to perform actions on packagings.\n
    Alias for the command is pack.
    """
    ctx.obj = ModelPackagingClient(url, token)


@packaging.command()
@click.option('--pack-id', '--id', help='Model packaging ID')
@click.option('--output-format', '-o', 'output_format', help='Output format',
              default=DEFAULT_OUTPUT_FORMAT, callback=validate_output_format)
@pass_obj
def get(client: ModelPackagingClient, pack_id: str, output_format: str):
    """
    Get packagings.\n
    The command without id argument retrieve all packagings.\n
    Get all packagings in json format:\n
        odahuflowctl pack get --format json\n
    Get packaging with "git-repo" id:\n
        odahuflowctl pack get --id git-repo\n
    Using jsonpath:\n
        odahuflowctl pack get -o 'jsonpath=[*].spec.reference'
    \f
    :param client: Model packaging HTTP client
    :param pack_id: Model packaging ID
    :param output_format: Output format
    :return:
    """
    packs = [client.get(pack_id)] if pack_id else client.get_all()

    format_output(packs, output_format)


@packaging.command()
@click.option('--pack-id', '--id', help='Model packaging ID')
@click.option('--file', '-f', type=click.Path(), required=True, help='Path to the file with packaging')
@click.option('--wait/--no-wait', default=True,
              help='no wait until scale will be finished')
@click.option('--artifact-name', type=str, help='Override artifact name from file')
@click.option('--timeout', default=DEFAULT_PACKAGING_TIMEOUT, type=int,
              help='timeout in seconds. for wait (if no-wait is off)')
@pass_obj
def create(client: ModelPackagingClient, pack_id: str, file: str, wait: bool, timeout: int,
           artifact_name: str):
    """
    Create a packaging.\n
    You should specify a path to file with a packaging. The file must contain only one packaging.
    For now, CLI supports yaml and JSON file formats.
    If you want to create multiples packagings than you should use "odahuflowctl res apply" instead.
    If you provide the packaging id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl pack create -f pack.yaml --id examples-git
    \f
    :param timeout: timeout in seconds. for wait (if no-wait is off)
    :param wait: no wait until scale will be finished
    :param client: Model packaging HTTP client
    :param pack_id: Model packaging ID
    :param file: Path to the file with only one packaging
    :param artifact_name: Override artifact name from file
    """
    pack = parse_resources_file_with_one_item(file).resource
    if not isinstance(pack, ModelPackaging):
        raise ValueError(f'Model packaging expected, but {type(pack)} provided')

    if pack_id:
        pack.id = pack_id

    if artifact_name:
        pack.spec.artifact_name = artifact_name

    mp = client.create(pack)
    click.echo(f"Start packing: {mp}")

    wait_packaging_finish(timeout, wait, pack.id, client)


@packaging.command()
@click.option('--pack-id', '--id', help='Model packaging ID')
@click.option('--file', '-f', type=click.Path(), required=True, help='Path to the file with packaging')
@click.option('--wait/--no-wait', default=True,
              help='no wait until scale will be finished')
@click.option('--artifact-name', type=str, help='Override artifact name from file')
@click.option('--timeout', default=DEFAULT_PACKAGING_TIMEOUT, type=int,
              help='timeout in seconds. for wait (if no-wait is off)')
@pass_obj
def edit(client: ModelPackagingClient, pack_id: str, file: str, wait: bool, timeout: int,
         artifact_name: str):
    """
    Update a packaging.\n
    You should specify a path to file with a packaging. The file must contain only one packaging.
    For now, CLI supports yaml and JSON file formats.
    If you want to update multiples packagings than you should use "odahuflowctl res apply" instead.
    If you provide the packaging id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl pack update -f pack.yaml --id examples-git
    \f
    :param client: Model packaging HTTP client
    :param pack_id: Model packaging ID
    :param file: Path to the file with only one packaging
    :param timeout: timeout in seconds. for wait (if no-wait is off)
    :param wait: no wait until scale will be finished
    :param artifact_name: Override artifact name from file
    """
    pack = parse_resources_file_with_one_item(file).resource
    if not isinstance(pack, ModelPackaging):
        raise ValueError(f'Model packaging expected, but {type(pack)} provided')

    if pack_id:
        pack.id = pack_id

    if artifact_name:
        pack.spec.artifact_name = artifact_name

    mp = client.edit(pack)
    click.echo(f"Rerun packing: {mp}")

    wait_packaging_finish(timeout, wait, pack.id, client)


@packaging.command()
@click.option('--pack-id', '--id', help='Model packaging ID')
@click.option('--file', '-f', type=click.Path(), help='Path to the file with packaging')
@click.option('--ignore-not-found/--not-ignore-not-found', default=False,
              help='ignore if Model Packaging is not found')
@pass_obj
def delete(client: ModelPackagingClient, pack_id: str, file: str, ignore_not_found: bool):
    """
    Delete a packaging.\n
    For this command, you must provide a packaging ID or path to file with one packaging.
    The file must contain only one packaging.
    If you want to delete multiples packagings than you should use "odahuflowctl res delete" instead.
    For now, CLI supports yaml and JSON file formats.
    The command will be failed if you provide both arguments.\n
    Usage example:\n
        * odahuflowctl pack delete --id examples-git\n
        * odahuflowctl pack delete -f pack.yaml
    \f
    :param client: Model packaging HTTP client
    :param pack_id: Model packaging ID
    :param file: Path to the file with only one packaging
    :param ignore_not_found: ignore if Model Packaging is not found
    """
    check_id_or_file_params_present(pack_id, file)

    if file:
        pack = parse_resources_file_with_one_item(file).resource
        if not isinstance(pack, ModelPackaging):
            raise ValueError(f'Model packaging expected, but {type(pack)} provided')

        pack_id = pack.id

    try:
        message = client.delete(pack_id)
        click.echo(message)
    except WrongHttpStatusCode as e:
        if e.status_code != 404 or not ignore_not_found:
            raise e

        click.echo(f'Model packaging {pack_id} was not found. Ignore')


@packaging.command()
@click.option('--pack-id', '--id', help='Model packaging ID')
@click.option('--file', '-f', type=click.Path(), help='Path to the file with packaging')
@click.option('--follow/--not-follow', default=True,
              help='Follow logs stream')
@pass_obj
def logs(client: ModelPackagingClient, pack_id: str, file: str, follow: bool):
    """
    Stream packaging logs.\n
    For this command, you must provide a packaging ID or path to file with one packaging.
    The file must contain only one packaging.
    If you want to delete multiples packagings than you should use "odahuflowctl res delete" instead.
    For now, CLI supports yaml and JSON file formats.
    The command will be failed if you provide both arguments.\n
    Usage example:\n
        * odahuflowctl pack delete --id examples-git\n
        * odahuflowctl pack delete -f pack.yaml
    \f
    :param follow: Follow logs stream
    :param client: Model packaging HTTP client
    :param pack_id: Model packaging ID
    :param file: Path to the file with only one packaging
    """
    if not pack_id and not file:
        raise ValueError(f'You should provide a packaging ID or file parameter, not both.')

    if pack_id and file:
        raise ValueError(f'You should provide a packaging ID or file parameter, not both.')

    if file:
        pack = parse_resources_file_with_one_item(file).resource
        if not isinstance(pack, ModelPackaging):
            raise ValueError(f'Model packaging expected, but {type(pack)} provided')

        pack_id = pack.id

    for msg in client.log(pack_id, follow):
        print_logs(msg)


def wait_packaging_finish(timeout: int, wait: bool, mp_id: str, mp_client: ModelPackagingClient):
    """
    Wait packaging to finish according command line arguments

    :param wait:
    :param timeout:
    :param mp_id: Model Packaging name
    :param mp_client: Model Packaging Client
    """
    if not wait:
        return

    start = time.time()
    if timeout <= 0:
        raise Exception('Invalid --timeout argument: should be positive integer')

    # We create a separate client for logs because it has the different timeout settings
    log_mp_client = ModelPackagingClient.construct_from_other(mp_client)
    log_mp_client.timeout = mp_client.timeout, LOG_READ_TIMEOUT_SECONDS

    click.echo("Logs streaming...")

    while True:
        elapsed = time.time() - start
        if elapsed > timeout:
            raise Exception(TIMEOUT_ERROR_MESSAGE)

        try:
            mp = mp_client.get(mp_id)
            if mp.status.state == SUCCEEDED_STATE:
                click.echo(f'Model {mp_id} was packed. Packaging took {round(time.time() - start)} seconds')
                return
            elif mp.status.state == FAILED_STATE:
                raise Exception(f'Model packaging {mp_id} was failed.')
            elif mp.status.state == "":
                click.echo(f"Can't determine the state of {mp.id}. Sleeping...")
            else:
                for msg in log_mp_client.log(mp.id, follow=True):
                    print_logs(msg)

        except (WrongHttpStatusCode, HTTPException, RequestException, APIConnectionException) as e:
            LOGGER.info('Callback have not confirmed completion of the operation. Exception: %s', str(e))

        LOGGER.debug('Sleep before next request')
        time.sleep(DEFAULT_WAIT_TIMEOUT)



#
#    Copyright 2020 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Test in this module must be applied for all entity commands
"""
import http
import json
import pathlib
import typing

import pytest
from click.testing import CliRunner
from odahuflow.cli.parsers.connection import IGNORE_NOT_FOUND_ERROR_MESSAGE, \
    ID_AND_FILE_MISSED_ERROR_MESSAGE
from odahuflow.cli.utils.output import JSON_OUTPUT_FORMAT, \
    JSONPATH_OUTPUT_FORMAT
from odahuflow.sdk.clients.api import WrongHttpStatusCode
from pytest_mock import MockFixture

from .data import ENTITY_ID, EntityTestData, ROUTER, DEPLOYMENT, PACKAGING_INTEGRATION, PACKAGING, TOOLCHAIN, TRAINING, \
    CONNECTION, generate_entities_for_test

ENTITY_TEST_DATA: typing.Dict[str, EntityTestData] = {
    "connection": CONNECTION,
    "training": TRAINING,
    "toolchain": TOOLCHAIN,
    "packaging": PACKAGING,
    "packaging_integration": PACKAGING_INTEGRATION,
    "deployments": DEPLOYMENT,
    "routes": ROUTER,
}

WRONG_OUTPUT_FORMAT = 'wrong-format'


def pytest_generate_tests(metafunc):
    generate_entities_for_test(metafunc, list(ENTITY_TEST_DATA.keys()))


@pytest.fixture
def entity_test_data(request) -> EntityTestData:
    return ENTITY_TEST_DATA[request.param]


def test_get(mocker: MockFixture, cli_runner: CliRunner, entity_test_data: EntityTestData):
    client_mock = mocker.patch.object(
        entity_test_data.entity_client.__class__, 'get',
        return_value=entity_test_data.entity,
    )

    result = cli_runner.invoke(entity_test_data.click_group, [
        'get', '--id', ENTITY_ID, '-o', JSON_OUTPUT_FORMAT,
    ], obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with(ENTITY_ID)
    assert result.exit_code == 0
    assert json.loads(result.output) == [entity_test_data.entity.to_dict()]


def test_get_all(mocker: MockFixture, cli_runner: CliRunner,
                 entity_test_data: EntityTestData):
    client_mock = mocker.patch.object(
        entity_test_data.entity_client.__class__,
        'get_all',
        return_value=[entity_test_data.entity],
    )

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['get', '-o', JSON_OUTPUT_FORMAT],
                               obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with()
    assert result.exit_code == 0
    assert json.loads(result.output) == [entity_test_data.entity.to_dict()]


def test_get_jsonpath(mocker: MockFixture, cli_runner: CliRunner,
                      entity_test_data: EntityTestData):
    client_mock = mocker.patch.object(
        entity_test_data.entity_client.__class__,
        'get',
        return_value=entity_test_data.entity,
    )

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['get', '--id', ENTITY_ID, '-o',
                                f'{JSONPATH_OUTPUT_FORMAT}=[*].id'],
                               obj=entity_test_data.entity_client)
    client_mock.assert_called_once_with(ENTITY_ID)
    assert result.exit_code == 0
    assert result.output.strip() == ENTITY_ID


def test_get_default_output_format(mocker: MockFixture, cli_runner: CliRunner,
                                   entity_test_data: EntityTestData):
    client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                      'get',
                                      return_value=entity_test_data.entity)

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['get', '--id', ENTITY_ID],
                               obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with(ENTITY_ID)
    assert result.exit_code == 0
    assert ENTITY_ID in result.stdout


def test_get_wrong_output_format(cli_runner: CliRunner,
                                 entity_test_data: EntityTestData):
    result = cli_runner.invoke(entity_test_data.click_group,
                               ['get', '--id', ENTITY_ID, '-o', WRONG_OUTPUT_FORMAT],
                               obj=entity_test_data.entity_client)

    assert result.exit_code != 0
    assert f'invalid choice: {WRONG_OUTPUT_FORMAT}' in result.output


def test_edit_wrong_kind(tmp_path: pathlib.Path, cli_runner: CliRunner,
                         entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps({**entity_test_data.entity.to_dict(), **{'kind': 'Wrong'}}))

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['edit', '-f', entity_file],
                               obj=entity_test_data.entity_client)

    assert result.exit_code != 0
    assert "Unknown kind of object: 'Wrong'" in str(result.exception)


def test_create_wrong_kind(tmp_path: pathlib.Path, cli_runner: CliRunner,
                           entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps({**entity_test_data.entity.to_dict(), **{'kind': 'Wrong'}}))

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['create', '-f', entity_file],
                               obj=entity_test_data.entity_client)

    assert result.exit_code != 0
    assert "Unknown kind of object: 'Wrong'" in str(result.exception)


def test_delete_id_and_file_present(tmp_path, cli_runner: CliRunner,
                                    entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps({**entity_test_data.entity.to_dict(), **{'kind': entity_test_data.kind}}))
    result = cli_runner.invoke(entity_test_data.click_group,
                               ['delete', '--id', 'some-id', '-f', entity_file],
                               obj=entity_test_data.entity_client)

    assert result.exit_code != 0
    assert ID_AND_FILE_MISSED_ERROR_MESSAGE in str(result.exception)


def test_delete_wrong_kind(tmp_path: pathlib.Path, cli_runner: CliRunner,
                           entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(json.dumps({**entity_test_data.entity.to_dict(), **{'kind': 'Wrong'}}))

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['delete', '-f', entity_file],
                               obj=entity_test_data.entity_client)

    assert result.exit_code != 0
    assert "Unknown kind of object: 'Wrong'" in str(result.exception)


def test_delete_ignore_not_found_enabled(mocker: MockFixture, cli_runner: CliRunner,
                                         entity_test_data: EntityTestData):
    client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                      'delete',
                                      side_effect=WrongHttpStatusCode(
                                          http.HTTPStatus.NOT_FOUND))

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['delete', '--id', ENTITY_ID,
                                '--ignore-not-found'],
                               obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with(ENTITY_ID)
    assert result.exit_code == 0
    assert IGNORE_NOT_FOUND_ERROR_MESSAGE.format(ENTITY_ID) in result.stdout


def test_delete_ignore_not_found_disabled(mocker: MockFixture, cli_runner: CliRunner,
                                          entity_test_data: EntityTestData):
    client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                      'delete',
                                      side_effect=WrongHttpStatusCode(
                                          http.HTTPStatus.NOT_FOUND))

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['delete', '--id', ENTITY_ID],
                               obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with(ENTITY_ID)
    assert result.exit_code != 0
    assert "Got error from server" in str(result.exception)


def test_delete_ignore_not_found_enabled_http_code(
        mocker: MockFixture, cli_runner: CliRunner, entity_test_data: EntityTestData,
):
    client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                      'delete',
                                      side_effect=WrongHttpStatusCode(
                                          http.HTTPStatus.BAD_REQUEST))

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['delete', '--id', ENTITY_ID,
                                '--ignore-not-found'],
                               obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with(ENTITY_ID)
    assert result.exit_code != 0
    assert "Got error from server" in str(result.exception)



#
#    Copyright 2020 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import json
import pathlib
import typing

import pytest
from click.testing import CliRunner
from pytest_mock import MockFixture

from .data import EntityTestData, generate_entities_for_test, TRAINING, PACKAGING, DEPLOYMENT, ROUTER

ENTITY_TEST_DATA: typing.Dict[str, EntityTestData] = {
    "training": TRAINING,
    "packaging": PACKAGING,
    "deployments": DEPLOYMENT,
    "routes": ROUTER,
}


def pytest_generate_tests(metafunc):
    generate_entities_for_test(metafunc, list(ENTITY_TEST_DATA.keys()))


@pytest.fixture
def entity_test_data(request) -> EntityTestData:
    return ENTITY_TEST_DATA[request.param]


def test_edit(tmp_path: pathlib.Path, mocker: MockFixture, cli_runner: CliRunner,
              entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps({**entity_test_data.entity.to_dict(), **{'kind': entity_test_data.kind}}))
    edit_client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                           'edit',
                                           return_value=entity_test_data.entity)
    get_client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                          'get',
                                          return_value=entity_test_data.entity)

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['edit', '-f', entity_file],
                               obj=entity_test_data.entity_client)

    assert result.exit_code == 0
    edit_client_mock.assert_called_once_with(entity_test_data.entity)
    get_client_mock.assert_called_once()


def test_edit_no_wait(mocker: MockFixture, tmp_path, cli_runner: CliRunner,
                      entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps(
            {**entity_test_data.entity.to_dict(), **{'kind': entity_test_data.kind}}))
    edit_client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                           'edit',
                                           return_value=entity_test_data.entity)
    get_client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                          'get',
                                          return_value=entity_test_data.entity)

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['edit', '-f', entity_file, '--no-wait'],
                               obj=entity_test_data.entity_client)

    assert result.exit_code == 0
    edit_client_mock.assert_called_once_with(entity_test_data.entity)
    get_client_mock.assert_not_called()


def test_create(tmp_path: pathlib.Path, mocker: MockFixture, cli_runner: CliRunner,
                entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps({**entity_test_data.entity.to_dict(), **{'kind': entity_test_data.kind}}))
    edit_client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                           'create',
                                           return_value=entity_test_data.entity)
    get_client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                          'get',
                                          return_value=entity_test_data.entity)

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['create', '-f', entity_file],
                               obj=entity_test_data.entity_client)

    assert result.exit_code == 0
    edit_client_mock.assert_called_once_with(entity_test_data.entity)
    get_client_mock.assert_called_once()


def test_create_no_wait(mocker: MockFixture, tmp_path, cli_runner: CliRunner,
                        entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps(
            {**entity_test_data.entity.to_dict(), **{'kind': entity_test_data.kind}}))
    edit_client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                           'create',
                                           return_value=entity_test_data.entity)
    get_client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                          'get',
                                          return_value=entity_test_data.entity)

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['create', '-f', entity_file, '--no-wait'],
                               obj=entity_test_data.entity_client)

    assert result.exit_code == 0
    edit_client_mock.assert_called_once_with(entity_test_data.entity)
    get_client_mock.assert_not_called()



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
EDGE commands for odahuflow cli
"""
import json

import click
from odahuflow.sdk import config
from odahuflow.sdk.clients.model import ModelClient, calculate_url


@click.group()
def model():
    """
    Allow you to perform actions on deployed models
    """
    pass


@model.command()
@click.option('--model-route', '--mr', default=config.MODEL_ROUTE_NAME, type=str, help='Name of Model Route')
@click.option('--model-deployment', '--md', default=config.MODEL_DEPLOYMENT_NAME, type=str,
              help='Name of Model Deployment')
@click.option('--url', default=config.MODEL_SERVER_URL, type=str, help='Full url of model server')
@click.option('--url-prefix', type=str, help='Url prefix of model server')
@click.option('--host', type=str, default=config.MODEL_HOST, help='Host of edge')
@click.option('--jwt', type=str, default=config.API_TOKEN, help='Model jwt token')
@click.option('--json', 'json_input', type=str, help='Json parameter. For example: --json {"x": 2}')
@click.option('--json-file', '--file', type=click.Path(exists=True), help='Path to json file')
def invoke(json_input, model_route: str, model_deployment: str, url: str, url_prefix: str,
           host: str, jwt: str, json_file: str):
    """
    Invoke model endpoint

    :param json_input:
    :param client:
    :return: None
    """
    if json_file:
        with open(json_file) as f:
            json_input = f.read()

    client = ModelClient(calculate_url(host, url, model_route, model_deployment, url_prefix), jwt)

    result = client.invoke(**json.loads(json_input))

    click.echo(json.dumps(result))


@model.command()
@click.option('--model-route', '--mr', default=config.MODEL_ROUTE_NAME, type=str, help='Name of Model Route')
@click.option('--model-deployment', '--md', default=config.MODEL_DEPLOYMENT_NAME, type=str,
              help='Name of Model Deployment')
@click.option('--url', default=config.MODEL_SERVER_URL, type=str, help='Full url of model server')
@click.option('--url-prefix', type=str, help='Url prefix of model server')
@click.option('--host', type=str, default=config.MODEL_HOST, help='Host of edge')
@click.option('--jwt', type=str, default=config.API_TOKEN, help='Model jwt token')
def info(model_route: str, model_deployment: str, url: str, url_prefix: str,
         host: str, jwt: str):
    """
    Get model information

    :param client: Model HTTP Client
    """
    client = ModelClient(calculate_url(host, url, model_route, model_deployment, url_prefix), jwt)

    result = client.info()

    print(result)



#
#    Copyright 2020 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Group of local commands
"""
import click
from odahuflow.cli.parsers.local.packaging import packaging
from odahuflow.cli.parsers.local.training import training
from odahuflow.cli.utils.abbr import AbbreviationGroup


@click.group(cls=AbbreviationGroup)
def local():
    """
    Train and package locally
    """
    pass


LOCAL_GROUPS = [
    training,
    packaging,
]

# Initialize local groups
for group in LOCAL_GROUPS:
    local.add_command(group)



#
#    Copyright 2020 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import logging
import sys
from contextlib import contextmanager
from typing import Optional

import click
from odahuflow.sdk.logger import is_verbose_enabled

LOG = logging.getLogger(__name__)

TIMEOUT_ERROR_MESSAGE = 'Time out: operation has not been confirmed'
IGNORE_NOT_FOUND_ERROR_MESSAGE = '{} was not found. Ignore'
ID_AND_FILE_MISSED_ERROR_MESSAGE = 'You should provide a ID or ' \
                                   'file parameter, not both.'


def check_id_or_file_params_present(
        entity_id: Optional[str], file: Optional[str]
) -> None:
    """
    Verify that only one parameter is None or empty
    """
    if not entity_id and not file:
        raise ValueError(ID_AND_FILE_MISSED_ERROR_MESSAGE)

    if entity_id and file:
        raise ValueError(ID_AND_FILE_MISSED_ERROR_MESSAGE)


@contextmanager
def cli_error_handler():
    """
    Print error message to stdout and exit if any error occurs
    """
    try:
        yield
    except Exception as e:
        # If the verbose flag is enabled than the stacktrace will be logged.
        LOG.exception('Exception occurs during CLI invocation')
        # This message always appears in stdout.
        # The verbose flag does not affect this.
        click.echo(f'Error: {str(e)}')

        if not is_verbose_enabled():
            click.echo('For more information rerun command with --verbose flag')

        # We can't reraise, for example, click.ClickException.
        # Because this function doesn't executes in click context,
        # so click library doesn't handle the exception.
        sys.exit(1)



#
#    Copyright 2020 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import json
import pathlib
import typing

import pytest
from click.testing import CliRunner
from pytest_mock import MockFixture

from .data import ENTITY_ID, EntityTestData, ROUTER, PACKAGING_INTEGRATION, PACKAGING, TOOLCHAIN, TRAINING, \
    CONNECTION, generate_entities_for_test

ENTITY_TEST_DATA: typing.Dict[str, EntityTestData] = {
    "connection": CONNECTION,
    "training": TRAINING,
    "toolchain": TOOLCHAIN,
    "packaging": PACKAGING,
    "packaging_integration": PACKAGING_INTEGRATION,
    "routes": ROUTER,
}


def pytest_generate_tests(metafunc):
    generate_entities_for_test(metafunc, list(ENTITY_TEST_DATA.keys()))


@pytest.fixture
def entity_test_data(request) -> EntityTestData:
    return ENTITY_TEST_DATA[request.param]


def test_delete_by_file(tmp_path: pathlib.Path, mocker: MockFixture, cli_runner: CliRunner,
                        entity_test_data: EntityTestData):
    message = "was deleted"
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps(
            {**entity_test_data.entity.to_dict(), **{'kind': entity_test_data.kind}}))
    client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                      'delete',
                                      return_value=message)

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['delete', '-f', entity_file],
                               obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with(entity_test_data.entity.id)
    assert result.exit_code == 0
    assert message in result.stdout


def test_delete_by_id(mocker: MockFixture, cli_runner: CliRunner,
                      entity_test_data: EntityTestData):
    message = "was deleted"
    client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                      'delete',
                                      return_value=message)

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['delete', '--id', ENTITY_ID],
                               obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with(ENTITY_ID)
    assert result.exit_code == 0
    assert message in result.stdout



#
#    Copyright 2020 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import json
import pathlib
import typing

import pytest
from pytest_mock import MockFixture
from click.testing import CliRunner
from odahuflow.cli.utils.output import JSON_OUTPUT_FORMAT, \
    JSONPATH_OUTPUT_FORMAT

from .data import ENTITY_ID, EntityTestData, PACKAGING_INTEGRATION, TOOLCHAIN, CONNECTION, \
    generate_entities_for_test

WRONG_OUTPUT_FORMAT = 'wrong-format'

ENTITY_TEST_DATA: typing.Dict[str, EntityTestData] = {
    "connection": CONNECTION,
    "toolchain": TOOLCHAIN,
    "packaging_integration": PACKAGING_INTEGRATION,
}


def pytest_generate_tests(metafunc):
    generate_entities_for_test(metafunc, list(ENTITY_TEST_DATA.keys()))


@pytest.fixture
def entity_test_data(request) -> EntityTestData:
    return ENTITY_TEST_DATA[request.param]


def test_edit(tmp_path: pathlib.Path, mocker: MockFixture, cli_runner: CliRunner,
              entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps({**entity_test_data.entity.to_dict(), **{'kind': entity_test_data.kind}}))
    client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                      'edit',
                                      return_value=entity_test_data.entity)

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['edit', '-f', entity_file, '-o',
                                JSON_OUTPUT_FORMAT],
                               obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with(entity_test_data.entity)
    assert result.exit_code == 0
    assert json.loads(result.output) == [entity_test_data.entity.to_dict()]


def test_edit_jsonpath(mocker: MockFixture, tmp_path, cli_runner: CliRunner,
                       entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps(
            {**entity_test_data.entity.to_dict(), **{'kind': entity_test_data.kind}}))
    client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                      'edit',
                                      return_value=entity_test_data.entity)

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['edit', '-f', entity_file, '-o',
                                f'{JSONPATH_OUTPUT_FORMAT}=[*].id'],
                               obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with(entity_test_data.entity)
    assert result.exit_code == 0
    assert result.output.strip() == ENTITY_ID


def test_edit_default_output_format(mocker: MockFixture, tmp_path, cli_runner: CliRunner,
                                    entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps(
            {**entity_test_data.entity.to_dict(), **{'kind': entity_test_data.kind}}))
    client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                      'edit',
                                      return_value=entity_test_data.entity)

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['edit', '-f', entity_file],
                               obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with(entity_test_data.entity)
    assert result.exit_code == 0
    assert ENTITY_ID in result.stdout


def test_edit_wrong_output_format(cli_runner: CliRunner,
                                  entity_test_data: EntityTestData):
    result = cli_runner.invoke(entity_test_data.click_group,
                               ['edit', '--id', ENTITY_ID, '-o', WRONG_OUTPUT_FORMAT],
                               obj=entity_test_data.entity_client)

    assert result.exit_code != 0
    assert f'invalid choice: {WRONG_OUTPUT_FORMAT}' in result.output


def test_create(tmp_path: pathlib.Path, mocker: MockFixture, cli_runner: CliRunner,
                entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps(
            {**entity_test_data.entity.to_dict(), **{'kind': entity_test_data.kind}}))
    client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                      'create',
                                      return_value=entity_test_data.entity)

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['create', '-f', entity_file, '-o',
                                JSON_OUTPUT_FORMAT],
                               obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with(entity_test_data.entity)
    assert result.exit_code == 0
    assert json.loads(result.output) == [entity_test_data.entity.to_dict()]


def test_create_wrong_output_format(cli_runner: CliRunner,
                                    entity_test_data: EntityTestData):
    result = cli_runner.invoke(entity_test_data.click_group,
                               ['edit', '--id', ENTITY_ID, '-o', WRONG_OUTPUT_FORMAT],
                               obj=entity_test_data.entity_client)

    assert result.exit_code != 0
    assert f'invalid choice: {WRONG_OUTPUT_FORMAT}' in result.output


def test_create_jsonpath(mocker: MockFixture, tmp_path, cli_runner: CliRunner,
                         entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps(
            {**entity_test_data.entity.to_dict(), **{'kind': entity_test_data.kind}}))
    client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                      'create',
                                      return_value=entity_test_data.entity)

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['create', '-f', entity_file, '-o',
                                f'{JSONPATH_OUTPUT_FORMAT}=[*].id'],
                               obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with(entity_test_data.entity)
    assert result.exit_code == 0
    assert result.output.strip() == ENTITY_ID


def test_create_default_output_format(mocker: MockFixture, tmp_path, cli_runner: CliRunner,
                                      entity_test_data: EntityTestData):
    entity_file = tmp_path / "entity.yaml"
    entity_file.write_text(
        json.dumps(
            {**entity_test_data.entity.to_dict(), **{'kind': entity_test_data.kind}}))
    client_mock = mocker.patch.object(entity_test_data.entity_client.__class__,
                                      'create',
                                      return_value=entity_test_data.entity)

    result = cli_runner.invoke(entity_test_data.click_group,
                               ['create', '-f', entity_file],
                               obj=entity_test_data.entity_client)

    client_mock.assert_called_once_with(entity_test_data.entity)
    assert result.exit_code == 0
    assert ENTITY_ID in result.stdout



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Config commands for odahuflow cli
"""
import sys

import click
from odahuflow.sdk import config


@click.group(name='config')
def config_group():
    """
    Odahuflow CLI config manipulation.\n
    Alias for the command is conf.
    """
    pass


@config_group.command(name="set")
@click.argument('key', type=str, required=True)
@click.argument('value', type=str, required=True)
def config_set(key: str, value: str):
    """
    Set configuration variable.\n
    Usage example:\n
        * odahuflowctl config set API_URL http://localhost:5000
    \f
    :param key: Configuration variable name
    :param value: Configuration variable name
    """
    variable_name = key.upper()
    _check_variable_exists_or_exit(variable_name)

    if not config.ALL_VARIABLES[variable_name].configurable_manually:
        raise Exception('Variable {} is not configurable manually'.format(variable_name))

    config.update_config_file(**{variable_name: value})

    _print_variable_information(variable_name, True)


@config_group.command(name="get")
@click.argument('key', type=str, required=True)
@click.option('--show-secrets/--no-show-secrets', default=False, help='Show tokens and passwords')
def config_get(key: str, show_secrets: bool):
    """
    Get configuration variable value\n
    Usage example:\n
        * odahuflowctl config get API_URL --show-secrets
    \f
    :param key: Configuration variable name
    :param show_secrets: Show tokens and passwords
    """
    variable_name = key.upper()
    _check_variable_exists_or_exit(variable_name)

    _print_variable_information(variable_name, show_secrets)


@config_group.command(name="all")
@click.option('--show-secrets/--no-show-secrets', default=False, help='Show tokens and passwords')
@click.option('--with-system/--no-with-system', default=False, help='Show with system variables')
def config_get_all(show_secrets: bool, with_system: bool):
    """
    Get all configuration variables\n
    Usage example:\n
        * odahuflowctl config all --show-secrets --with-system
    \f
    :param with_system: Show with system variables
    :param show_secrets: Show tokens and passwords
    """
    configurable_values = filter(lambda i: i[1].configurable_manually, config.ALL_VARIABLES.items())
    non_configurable_values = filter(lambda i: not i[1].configurable_manually, config.ALL_VARIABLES.items())

    click.echo('Configurable manually variables:\n===========================')
    for name, _ in configurable_values:
        _print_variable_information(name, show_secrets)

    if with_system:
        click.echo('\n\n')

        click.echo('System variables:\n===========================')
        for name, _ in non_configurable_values:
            _print_variable_information(name, show_secrets)


@config_group.command(name="path")
def config_path():
    """
    Get configuration storage
    Usage example:\n
        * odahuflowctl config path
    \f
    """
    click.echo(config.get_config_file_path())


def _check_variable_exists_or_exit(name: str):
    """
    Check that variable with desired name exists or print error message and exit with code 1

    :param name: name of variable, case-sensitive
    """
    if name not in config.ALL_VARIABLES:
        click.echo('Variable {!r} is unknown'.format(name))
        sys.exit(1)


def _print_variable_information(name: str, show_secrets: bool = False):
    """
    Print information about variable to console

    :param name: name of variable, case-sensitive
    :param show_secrets: do not mask credentials
    :return: None
    """
    description = config.ALL_VARIABLES[name]
    current_value = getattr(config, name)
    is_secret = any(sub in name for sub in ('_PASSWORD', '_TOKEN'))
    click.echo('{} - {}\n  default: {!r}'.format(name, description.description, description.default))
    if current_value != description.default:
        if is_secret and not show_secrets:
            current_value = '****'
        click.echo('  current: {!r}'.format(current_value))



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import os
import sys

import click
from odahuflow.sdk.gppi.executor import GPPITrainedModelBinary

ODAHUFLOW_GPPI_MODEL_PATH_ENV_NAME = 'ODAHUFLOW_GPPI_MODEL_PATH'
ODAHUFLOW_CONDA_ENV_NAME = 'ODAHUFLOW_CONDA'


class GppiCommandContextObj:

    def __init__(self, model_binary: GPPITrainedModelBinary):
        self.model_binary = model_binary


@click.group()
@click.option('--gppi-model-path', '-m', type=click.Path(exists=True, file_okay=False),
              envvar=ODAHUFLOW_GPPI_MODEL_PATH_ENV_NAME)
@click.option('--use-current-env/--not-use-current-env', 'use_current_env',
              help='Use current environment', default=False)
@click.option('--env-name', '-e', 'env_name', help='Environment name to run GPPI model', type=click.STRING,
              envvar=ODAHUFLOW_CONDA_ENV_NAME)
@click.option('--skip-deps/--not-skip-deps', 'skip_deps', help='Not install dependencies', default=False)
@click.pass_context
def gppi(ctx, gppi_model_path: str, use_current_env: bool = False, env_name: str = '', skip_deps: bool = False):
    """
    Allow you to perform actions on odahuflow gppi models
    """

    if not gppi_model_path:
        click.echo(f'--gppi-model-path OR {ODAHUFLOW_GPPI_MODEL_PATH_ENV_NAME} env var must be provided')
        sys.exit(1)

    mb = GPPITrainedModelBinary(gppi_model_path, use_current_env, env_name, skip_deps)
    ctx.obj = GppiCommandContextObj(mb)


@gppi.command()
@click.pass_context
def test(ctx):
    """
    Initialize GPPI model and try to execute api

    Next API are tested:

    Executes .info() method

    Executes .predict_on_matrix() method
    with data deserialized from head_input.pkl as a parameter
    (only if head_input.pkl exists as a model artifact)
    """
    ctx_obj: GppiCommandContextObj = ctx.obj
    ctx_obj.model_binary.self_check()
    click.echo(f'OK\nGPPI is correct and could be packaged or deployed')


@gppi.command()
@click.argument('input-file', type=click.Path(exists=True, dir_okay=False))
@click.argument('output-dir', type=click.Path(exists=True, file_okay=False))
@click.option('--output-file-name', '-o', default='results.json')
@click.pass_context
def predict(ctx, input_file: str, output_dir: str, output_file_name: str):
    """

    :param ctx:
    :param input_file: Input JSON file for predictions
    :param output_dir: Output directory where results should be saved
    :param output_file_name: Output filename with predictions
    :return:
    """
    ctx_obj: GppiCommandContextObj = ctx.obj
    ctx_obj.model_binary.predict(input_file, output_dir, output_file_name)
    full_path = os.path.join(output_dir, output_file_name)
    click.echo(f'Prediction file: {full_path}')



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import http

import click
from odahuflow.cli.utils.client import pass_obj
from odahuflow.cli.utils.error_handler import ID_AND_FILE_MISSED_ERROR_MESSAGE, \
    IGNORE_NOT_FOUND_ERROR_MESSAGE
from odahuflow.cli.utils.output import format_output, DEFAULT_OUTPUT_FORMAT, validate_output_format
from odahuflow.sdk import config
from odahuflow.sdk.clients.connection import ConnectionClient
from odahuflow.sdk.clients.api import WrongHttpStatusCode
from odahuflow.sdk.clients.api_aggregated import parse_resources_file_with_one_item
from odahuflow.sdk.models import Connection


@click.group()
@click.option('--url', help='API server host', default=config.API_URL)
@click.option('--token', help='API server jwt token', default=config.API_TOKEN)
@click.pass_context
def connection(ctx: click.core.Context, url: str, token: str):
    """
    Allow you to perform actions on connections.\n
    Alias for the command is conn.
    """
    ctx.obj = ConnectionClient(url, token)


@connection.command()
@click.option('--conn-id', '--id', help='Connection ID')
@click.option('--output-format', '-o', 'output_format', help='Output format',
              default=DEFAULT_OUTPUT_FORMAT, callback=validate_output_format)
@click.option('--decrypted', '-d', help='Flag means that connection sensitive data should be decrypted',
              default=False, is_flag=True)
@pass_obj
def get(client: ConnectionClient, conn_id: str, output_format: str, decrypted: bool):
    """
    Get connections.\n
    The command without id argument retrieve all connections.\n
    Get all connections in json format:\n
        odahuflowctl conn get --format json\n
    Get connection with "git-repo" id:\n
        odahuflowctl conn get --id git-repo\n
    Using jsonpath:\n
        odahuflowctl conn get -o 'jsonpath=[*].spec.reference'
    \f
    :param decrypted: if set than decrypted connection will be returned
    :param client: Connection HTTP client
    :param conn_id: Connection ID
    :param output_format: Output format
    :return:
    """
    if conn_id:
        if decrypted:
            conn = client.get_decrypted(conn_id)
        else:
            conn = client.get(conn_id)

        conns = [conn]
    else:
        conns = client.get_all()

    format_output(conns, output_format)


@connection.command()
@click.option('--conn-id', '--id', help='Connection ID')
@click.option('--file', '-f', type=click.Path(), required=True, help='Path to the file with connection')
@click.option('--output-format', '-o', 'output_format', help='Output format',
              default=DEFAULT_OUTPUT_FORMAT, callback=validate_output_format)
@pass_obj
def create(client: ConnectionClient, conn_id: str, file: str, output_format: str):
    """
    Create a connection.\n
    You should specify a path to file with a connection. The file must contain only one connection.
    For now, CLI supports yaml and JSON file formats.
    If you want to create multiples connections than you should use "odahuflowctl res apply" instead.
    If you provide the connection id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl conn create -f conn.yaml --id examples-git
    \f
    :param client: Connection HTTP client
    :param conn_id: Connection ID
    :param file: Path to the file with only one connection
    :param output_format: Output format
    """
    conn = parse_resources_file_with_one_item(file).resource
    if not isinstance(conn, Connection):
        raise ValueError(f'Connection expected, but {type(conn)} provided')

    if conn_id:
        conn.id = conn_id

    click.echo(format_output([client.create(conn)], output_format))


@connection.command()
@click.option('--conn-id', '--id', help='Connection ID')
@click.option('--file', '-f', type=click.Path(), required=True, help='Path to the file with connection')
@click.option('--output-format', '-o', 'output_format', help='Output format',
              default=DEFAULT_OUTPUT_FORMAT, callback=validate_output_format)
@pass_obj
def edit(client: ConnectionClient, conn_id: str, file: str, output_format: str):
    """
    Update a connection.\n
    You should specify a path to file with a connection. The file must contain only one connection.
    For now, CLI supports yaml and JSON file formats.
    If you want to update multiples connections than you should use "odahuflowctl res apply" instead.
    If you provide the connection id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl conn update -f conn.yaml --id examples-git
    \f
    :param client: Connection HTTP client
    :param conn_id: Connection ID
    :param file: Path to the file with only one connection
    :param output_format: Output format
    """
    conn = parse_resources_file_with_one_item(file).resource
    if not isinstance(conn, Connection):
        raise ValueError(f'Connection expected, but {type(conn)} provided')

    if conn_id:
        conn.id = conn_id

    click.echo(format_output([client.edit(conn)], output_format))


@connection.command()
@click.option('--conn-id', '--id', help='Connection ID')
@click.option('--file', '-f', type=click.Path(), help='Path to the file with connection')
@click.option('--ignore-not-found/--not-ignore-not-found', default=False,
              help='ignore if connection is not found')
@pass_obj
def delete(client: ConnectionClient, conn_id: str, file: str, ignore_not_found: bool):
    """
    Delete a connection.\n
    For this command, you must provide a connection ID or path to file with one connection.
    The file must contain only one connection.
    If you want to delete multiples connections than you should use "odahuflowctl res delete" instead.
    For now, CLI supports yaml and JSON file formats.
    The command will be failed if you provide both arguments.\n
    Usage example:\n
        * odahuflowctl conn delete --id examples-git\n
        * odahuflowctl conn delete -f conn.yaml\n
        * odahuflowctl conn delete --id examples-git --ignore-not-found
    \f
    :param client: Connection HTTP client
    :param conn_id: Connection ID
    :param file: Path to the file with only one connection
    :param ignore_not_found: ignore if connection is not found
    """
    if not conn_id and not file:
        raise ValueError(ID_AND_FILE_MISSED_ERROR_MESSAGE)

    if conn_id and file:
        raise ValueError(ID_AND_FILE_MISSED_ERROR_MESSAGE)

    if file:
        conn = parse_resources_file_with_one_item(file).resource
        if not isinstance(conn, Connection):
            raise ValueError(f'Connection expected, but {type(conn)} provided')

        conn_id = conn.id

    try:
        click.echo(client.delete(conn_id))
    except WrongHttpStatusCode as e:
        if e.status_code != http.HTTPStatus.NOT_FOUND or not ignore_not_found:
            raise e

        click.echo(IGNORE_NOT_FOUND_ERROR_MESSAGE.format(conn_id))



#
#    Copyright 2020 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Local training commands for odahuflow cli
"""
import json
import logging
from typing import List, Dict, Optional

import click
from odahuflow.cli.utils.client import pass_obj
from odahuflow.cli.utils.output import PLAIN_TEXT_OUTPUT_FORMAT, JSON_OUTPUT_FORMAT
from odahuflow.sdk import config
from odahuflow.sdk.clients.api_aggregated import \
    parse_resources_file, \
    parse_resources_dir, OdahuflowCloudResourceUpdatePair
from odahuflow.sdk.clients.toolchain_integration import ToolchainIntegrationClient
from odahuflow.sdk.clients.training import ModelTraining, ModelTrainingClient
from odahuflow.sdk.local.training import start_train, list_local_trainings, cleanup_local_artifacts, \
    cleanup_training_docker_containers
from odahuflow.sdk.models import ToolchainIntegration, K8sTrainer

LOGGER = logging.getLogger(__name__)


@click.group()
@click.option('--url', help='API server host', default=config.API_URL)
@click.option('--token', help='API server jwt token', default=config.API_TOKEN)
@click.pass_context
def training(ctx: click.core.Context, url: str, token: str):
    """
    Local training process.\n
    Alias for the command is train.
    """
    ctx.obj = ModelTrainingClient(url, token)


@training.command("list")
@click.option('--output-format', '-o', 'output_format',
              default=PLAIN_TEXT_OUTPUT_FORMAT, type=click.Choice([PLAIN_TEXT_OUTPUT_FORMAT, JSON_OUTPUT_FORMAT]))
def training_list(output_format: str):
    """
    Get list of local training artifacts.\n
    Get all training artifacts:\n
        odahuflowctl local train list\n
    Get all training artifacts in json format:\n
        odahuflowctl local train -o json\n
    \f
    :param output_format: Output format
    """
    artifacts = list_local_trainings()

    if output_format == JSON_OUTPUT_FORMAT:
        click.echo(json.dumps(artifacts, indent=2))
    else:
        if not artifacts:
            click.echo('Artifacts not found')
            return

        click.echo('Training artifacts:')

        for artifact in artifacts:
            click.echo(f'* {artifact}')


@training.command('cleanup-artifacts')
def cleanup_artifacts():
    """
    Delete all training local artifacts.
    Usage example:\n
        * odahuflowctl local train cleanup\n
    \f
    """
    cleanup_local_artifacts()


@training.command('cleanup-containers')
def cleanup_containers():
    """
    Delete all training docker containers.
    Usage example:\n
        * odahuflowctl local train cleanup\n
    \f
    """
    cleanup_training_docker_containers()


@training.command()
@click.option('--train-id', '--id', help='Model training ID', required=True)
@click.option('--manifest-file', '-f', type=click.Path(), multiple=True,
              help='Path to a ODAHU-flow manifest file')
@click.option('--manifest-dir', '-d', type=click.Path(), multiple=True,
              help='Path to a directory with ODAHU-flow manifest files')
@click.option('--output-dir', '--output', type=click.Path(), help='Directory where model artifact will be saved')
@pass_obj
def run(client: ModelTrainingClient, train_id: str, manifest_file: List[str], manifest_dir: List[str],
        output_dir: str):
    """
    Start a training process locally.\n
    Usage example:\n
        * odahuflowctl local train run --id examples-git\n
    \f
    """
    entities: List[OdahuflowCloudResourceUpdatePair] = []
    for file_path in manifest_file:
        entities.extend(parse_resources_file(file_path).changes)

    for dir_path in manifest_dir:
        entities.extend(parse_resources_dir(dir_path))

    mt: Optional[ModelTraining] = None

    # find a training
    toolchains: Dict[str, ToolchainIntegration] = {}
    for entity in map(lambda x: x.resource, entities):
        if isinstance(entity, ToolchainIntegration):
            toolchains[entity.id] = entity
        elif isinstance(entity, ModelTraining) and entity.id == train_id:
            mt = entity

    if not mt:
        LOGGER.debug(f'{train_id} training not found. Trying to retrieve it from API server')
        mt = client.get(train_id)

    toolchain = toolchains.get(mt.spec.toolchain)
    if not toolchain:
        LOGGER.debug(f'{toolchain} toolchain not found. Trying to retrieve it from API server')
        toolchain = ToolchainIntegrationClient.construct_from_other(client).get(mt.spec.toolchain)

    trainer = K8sTrainer(
        model_training=mt,
        toolchain_integration=toolchain,
    )

    start_train(trainer, output_dir)



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Output util functions
"""

import json
import typing
from collections import OrderedDict

import click
import yaml
from jsonpath_rw import parse
from texttable import Texttable

from odahuflow.sdk.models.base_model_ import Model

SPLIT_SIGN = "="

MAX_TABLE_WIDTH = 200
MAX_TABLE_VALUE_LENGTH = 50

PLAIN_TEXT_OUTPUT_FORMAT = "text"
JSON_OUTPUT_FORMAT = "json"
YAML_OUTPUT_FORMAT = "yaml"
TABLE_OUTPUT_FORMAT = "table"
JSONPATH_OUTPUT_FORMAT = "jsonpath"
DEFAULT_OUTPUT_FORMAT = YAML_OUTPUT_FORMAT

ALL_OUTPUT_FORMATS = [
    JSON_OUTPUT_FORMAT,
    TABLE_OUTPUT_FORMAT,
    YAML_OUTPUT_FORMAT,
    JSONPATH_OUTPUT_FORMAT
]


# pylint: disable=W0613
def validate_output_format(ctx, param, value):
    """
    Custom validator for "--output-format" parameter.
    It's an analog of click.Choise

    :param ctx: click context
    :param param: param name
    :param value: param value
    :return: cli value
    """
    splited_output_format = value.split(SPLIT_SIGN, maxsplit=1)
    if splited_output_format[0] not in ALL_OUTPUT_FORMATS:
        raise click.BadParameter(f'invalid choice: {value}. (choose from {", ".join(ALL_OUTPUT_FORMATS)})')

    return value


def flatten_dict(d: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
    """
    Flat keys of a dictionary.
    For example: {"a": {"b": 1}, "c": 2} produces {"a.b": 1, "c": 2}
    :param d: input dictionary
    :return: flatten dictionary
    """

    def expand(key, value):
        if isinstance(value, dict):
            return [(key + '.' + k, v) for k, v in flatten_dict(value).items()]
        else:
            return [(key, value)]

    items = [item for k, v in d.items() for item in expand(k, v)]

    return OrderedDict(items)


def _restrict_table_value_length(value: str) -> str:
    """
    Reduce the length of a table cell
    :param value: table cell
    :return: reduced cell of a table
    """
    if len(value) < MAX_TABLE_VALUE_LENGTH:
        return value
    else:
        return str(value)[:MAX_TABLE_VALUE_LENGTH] + "..."


def _convert_model_values(model: Model) -> typing.List[str]:
    """
    Convert a odahuflow entity to a flatten dict
    :param model: odahuflow entity
    :return: flatten dict
    """
    return list(
        map(
            lambda x: _restrict_table_value_length(str(x)),
            flatten_dict(model.to_dict()).values()
        )
    )


def show_table(models: typing.List[Model]):
    """
    Produce table to stdout from list of odahuflow models
    :param models: odahuflow models
    """
    if not models:
        click.echo("Empty!")
        return

    headers = list(flatten_dict(models[0].to_dict()).keys())
    table = Texttable(MAX_TABLE_WIDTH)

    table.add_rows([headers] + [_convert_model_values(md) for md in models])
    click.echo(table.draw() + "\n")


def extract_jsonpath(models: typing.List[Model], expr: str):
    """
    Apply jsonpath expression to the list of modules
    :param models: odahuflow models
    :param expr: jsonpath expression
    """
    jsonpath_expr = parse(expr)

    result = [match.value for match in jsonpath_expr.find([md.to_dict() for md in models])]

    click.echo(result[0] if len(result) == 1 else json.dumps(result))


def format_output(models: typing.List[Model], output_format: str):
    """
    Write odahuflow models to stdout by specific output format
    :param models: odahuflow models
    :param output_format: output format
    """
    splitted_output_format = output_format.split(SPLIT_SIGN, maxsplit=1)
    output_format, expression = splitted_output_format[0], ""
    if len(splitted_output_format) == 2:
        expression = splitted_output_format[1]

    if output_format == JSON_OUTPUT_FORMAT:
        click.echo(json.dumps([m.to_dict() for m in models]))
    elif output_format == YAML_OUTPUT_FORMAT:
        click.echo(yaml.dump([m.to_dict() for m in models]))
    elif output_format == TABLE_OUTPUT_FORMAT:
        show_table(models)
    elif output_format == JSONPATH_OUTPUT_FORMAT:
        extract_jsonpath(models, expression)
    else:
        raise ValueError(f'Not expected {output_format} output format')



#
#    Copyright 2020 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
from click.testing import CliRunner
from odahuflow.cli.parsers import template


def test_get_all_template_names(cli_runner: CliRunner):
    result = cli_runner.invoke(template.template, ['all'])

    assert result.exit_code == 0
    assert result.output
    assert "* deployment\n" in result.output
    assert "* mlflow_training\n" in result.stdout
    assert "* gcp_connection\n" in result.stdout


def test_generate_template_by_name(cli_runner: CliRunner):
    result = cli_runner.invoke(template.template, ['generate', '--name', 'deployment'])

    assert result.exit_code == 0
    assert result.output
    assert "image: <model/image:tag>" in result.output
    assert "kind: ModelDeployment" in result.output


def test_generate_not_present_template(cli_runner: CliRunner):
    result = cli_runner.invoke(template.template, ['generate', '--name', 'not_present'])

    assert result.exit_code != 0
    assert result.exception
    assert "Cannot find not_present template" in str(result.exception)



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Version file
"""
import click
from odahuflow.sdk.version import __version__ as __sdk_version__

__version__ = '1.1.2'


@click.command()
def version():
    """
    Show version of cli and sdk
    """
    click.echo(f'cli version: {__version__}\n'
               f'sdk version: {__sdk_version__}')



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Template CLI commands
"""
import click
from odahuflow.sdk.clients.templates import get_odahuflow_template_content
from odahuflow.sdk.clients.templates import get_odahuflow_template_names


@click.group()
def template():
    """
    Allow you to perform actions on odahuflow template files.\n
    Alias for the command is temp.
    """
    pass


@template.command(name='all')
def template_all():
    """
    Get all template names.\n
    Usage example:\n
        * odahuflowctl template all\n
    """
    nl = "\n * "
    click.echo(f'Templates:{nl}{nl.join(get_odahuflow_template_names())}')


@template.command()
@click.option('--name', help='Template name', required=True)
def generate(name: str):
    """
    Generate a template by name.
    Format of templates is YAML.\n
    To find all templates execute the following command:\n
        * odahuflowctl template all\n
    Usage example:\n
        * odahuflowctl template generate --name deployment
    """
    click.echo(get_odahuflow_template_content(name))



# pylint: disable=W0613
import click


def positive_number(ctx, param: click.core.Option, value):
    """
    Check that CLI parameters is positive number

    :param ctx: click context
    :param param: param name
    :param value: pa-ram value
    :return: cli value
    """
    if int(value) <= 0:
        raise click.BadParameter(f'{param.name} must be positive integer')

    return value



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
API commands for odahuflow cli
"""
import logging
import sys
import typing

import click
from odahuflow.cli.utils.client import pass_obj
from odahuflow.sdk import config
from odahuflow.sdk.clients.api import RemoteAPIClient
from odahuflow.sdk.clients.api_aggregated import apply as api_aggregated_apply
from odahuflow.sdk.clients.api_aggregated import parse_resources_file, OdahuflowCloudResourceUpdatePair

LOGGER = logging.getLogger(__name__)


@click.group()
@click.option('--url', help='API server host', default=config.API_URL)
@click.option('--token', help='API server jwt token', default=config.API_TOKEN)
@click.pass_context
def bulk(ctx: click.core.Context, url: str, token: str):
    """
    Bulk operations on Odahuflow resources
    """
    ctx.obj = RemoteAPIClient(url, token)


@bulk.command()
@click.argument('file', required=True, type=click.Path())
@pass_obj
def apply(client: RemoteAPIClient, file: str):
    """
    Create/Update Odahuflow resources on an API.\n
    You should specify a path to file with resources.
    For now, CLI supports yaml and JSON file formats.
    Usage example:\n
        * odahuflowctl bulk apply resources.odahuflow.yaml
    \f
    :param client: Generic API HTTP client
    :param file: Path to the file with odahuflow resources
    """
    process_bulk_operation(client, file, False)


@bulk.command()
@click.argument('file', required=True, type=click.Path())
@pass_obj
def delete(client: RemoteAPIClient, file: str):
    """
    Remove Odahuflow resources from an API.\n
    You should specify a path to file with resources.
    For now, CLI supports yaml and JSON file formats.
    Usage example:\n
        * odahuflowctl bulk delete resources.odahuflow.yaml
    \f
    :param client: Generic API HTTP client
    :param file: Path to the file with odahuflow resources
    """
    process_bulk_operation(client, file, True)


def _print_resources_info_counter(objects: typing.Tuple[OdahuflowCloudResourceUpdatePair]) -> str:
    """
    Output count of resources with their's types

    :param objects: resources to print
    :return: str
    """
    names = ', '.join(f'{type(obj.resource).__name__} {obj.resource_id}' for obj in objects)
    if objects:
        return f'{len(objects)} ({names})'
    else:
        return str(len(objects))


def process_bulk_operation(api_client: RemoteAPIClient, filename: str, is_removal: bool):
    """
    Apply bulk operation helper

    :param api_client: base API client to extract connection options from
    :param filename: path to file with resources
    :param is_removal: is it removal operation
    """
    odahuflow_resources = parse_resources_file(filename)
    result = api_aggregated_apply(odahuflow_resources, api_client, is_removal)
    output = ['Operation completed']
    if result.created:
        output.append(f'created resources: {_print_resources_info_counter(result.created)}')
    if result.changed:
        output.append(f'changed resources: {_print_resources_info_counter(result.changed)}')
    if result.removed:
        output.append(f'removed resources: {_print_resources_info_counter(result.removed)}')
    click.echo(', '.join(output))

    if result.errors:
        click.echo(f'Some errors detected:')
        for error in result.errors:
            click.echo(f'\t{error}')
        sys.exit(1)



#
#    Copyright 2020 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Local packaging commands for odahuflow cli
"""
import logging
from typing import List, Dict, Optional

import click
from odahuflow.cli.utils.client import pass_obj
from odahuflow.sdk import config
from odahuflow.sdk.clients.api_aggregated import \
    parse_resources_file, \
    parse_resources_dir, OdahuflowCloudResourceUpdatePair
from odahuflow.sdk.clients.packaging import ModelPackagingClient
from odahuflow.sdk.clients.packaging_integration import PackagingIntegrationClient
from odahuflow.sdk.local.packaging import start_package, cleanup_packaging_docker_containers
from odahuflow.sdk.models import K8sPackager, ModelPackaging, PackagingIntegration

LOGGER = logging.getLogger(__name__)


@click.group()
@click.option('--url', help='API server host', default=config.API_URL)
@click.option('--token', help='API server jwt token', default=config.API_TOKEN)
@click.pass_context
def packaging(ctx: click.core.Context, url: str, token: str):
    """
    Local packaging process.\n
    Alias for the command is pack.
    """
    ctx.obj = ModelPackagingClient(url, token)


@packaging.command('cleanup-containers')
@pass_obj
def cleanup_containers():
    """
    Delete all packaging docker containers.
    Usage example:\n
        * odahuflowctl local pack cleanup\n
    \f
    """
    cleanup_packaging_docker_containers()


@packaging.command()
@click.option('--pack-id', '--id', help='Model packaging ID', required=True)
@click.option('--manifest-file', '-f', type=click.Path(), multiple=True,
              help='Path to a ODAHU-flow manifest file')
@click.option('--manifest-dir', '-d', type=click.Path(), multiple=True,
              help='Path to a directory with ODAHU-flow manifest files')
@click.option('--artifact-path', type=click.Path(),
              help='Path to a training artifact')
@click.option('--artifact-name', '-a', type=str, help='Override artifact name from file')
@click.option('--disable-package-targets/--no-disable-package-targets', 'is_target_disabled',
              default=True, help='Disable all targets in packaging')
@pass_obj
def run(client: ModelPackagingClient, pack_id: str, manifest_file: List[str], manifest_dir: List[str],
        artifact_path: str, artifact_name: str, is_target_disabled: bool):
    """
    Start a packaging process locally.\n
    Usage example:\n
        * odahuflowctl local pack run --id wine\n
    \f
    """
    entities: List[OdahuflowCloudResourceUpdatePair] = []
    for file_path in manifest_file:
        entities.extend(parse_resources_file(file_path).changes)

    for dir_path in manifest_dir:
        entities.extend(parse_resources_dir(dir_path))

    mp: Optional[ModelPackaging] = None

    packagers: Dict[str, PackagingIntegration] = {}
    for entity in map(lambda x: x.resource, entities):
        if isinstance(entity, PackagingIntegration):
            packagers[entity.id] = entity
        elif isinstance(entity, ModelPackaging) and entity.id == pack_id:
            mp = entity

    if not mp:
        LOGGER.debug(
            f'The {pack_id} packaging not found in the manifest files.'
            f' Trying to retrieve it from API server'
        )
        mp = client.get(pack_id)

    integration_name = mp.spec.integration_name
    packager = packagers.get(integration_name)
    if not packager:
        LOGGER.debug(
            f'The {integration_name} packager not found in the manifest files.'
            f' Trying to retrieve it from API server'
        )
        packager = PackagingIntegrationClient.construct_from_other(client).get(integration_name)

    if artifact_name:
        mp.spec.artifact_name = artifact_name

        LOGGER.debug('Override the artifact namesdsdsdsdsdsdsd')

    if is_target_disabled:
        mp.spec.targets = []

    k8s_packager = K8sPackager(
        model_packaging=mp,
        packaging_integration=packager,
        targets=[],
    )

    result = start_package(k8s_packager, artifact_path)

    click.echo('Packager results:')
    for key, value in result.items():
        click.echo(f'* {key} - {value}')



#
#    Copyright 2020 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Test data for entity CLI commands
"""
import typing

import click
from odahuflow.cli.parsers import connection, training, toolchain_integration, packaging, packaging_integration, \
    deployment, route
from odahuflow.sdk.clients.api import RemoteAPIClient
from odahuflow.sdk.clients.connection import ConnectionClient
from odahuflow.sdk.clients.deployment import ModelDeploymentClient, READY_STATE
from odahuflow.sdk.clients.packaging import ModelPackagingClient, SUCCEEDED_STATE
from odahuflow.sdk.clients.packaging_integration import PackagingIntegrationClient
from odahuflow.sdk.clients.route import ModelRouteClient
from odahuflow.sdk.clients.toolchain_integration import ToolchainIntegrationClient
from odahuflow.sdk.clients.training import ModelTrainingClient
from odahuflow.sdk.models import ModelTraining, Connection, ConnectionSpec, ModelTrainingSpec, ModelIdentity, \
    ToolchainIntegration, ToolchainIntegrationSpec, ModelPackaging, ModelPackagingSpec, PackagingIntegration, \
    PackagingIntegrationSpec, ModelDeployment, ModelDeploymentSpec, ModelRoute, ModelRouteSpec, \
    ModelTrainingStatus, ModelPackagingStatus, ModelDeploymentStatus, ModelRouteStatus
from odahuflow.sdk.models.base_model_ import Model

ENTITY_ID = 'entity-id'


# This interface for any test cases that tests CLI commands for entities
class EntityTestData(typing.NamedTuple):
    entity_client: RemoteAPIClient
    entity: Model
    click_group: click.Group
    # For example, ModelTraining or Connection
    kind: str


def generate_entities_for_test(metafunc, keys: typing.List[str]):
    if "entity_test_data" in metafunc.fixturenames:
        metafunc.parametrize(
            "entity_test_data",
            keys,
            indirect=True,
        )


CONNECTION = EntityTestData(
    ConnectionClient(),
    Connection(
        id=ENTITY_ID,
        spec=ConnectionSpec(
            key_secret="mock-key-secret",
            uri="mock-url",
        ),
    ),
    connection.connection,
    'Connection',
)

TRAINING = EntityTestData(
    ModelTrainingClient(),
    ModelTraining(
        id=ENTITY_ID,
        spec=ModelTrainingSpec(
            work_dir="/1/2/3",
            model=ModelIdentity(
                name="name",
                version="version"
            )
        ),
        status=ModelTrainingStatus(
            state=SUCCEEDED_STATE
        )
    ),
    training.training,
    'ModelTraining'
)

TOOLCHAIN = EntityTestData(
    ToolchainIntegrationClient(),
    ToolchainIntegration(
        id=ENTITY_ID,
        spec=ToolchainIntegrationSpec(
            default_image="mock-image",
            entrypoint="default-entrypoint",
        ),
    ),
    toolchain_integration.toolchain_integration,
    'ToolchainIntegration',
)

PACKAGING = EntityTestData(
    ModelPackagingClient(),
    ModelPackaging(
        id=ENTITY_ID,
        spec=ModelPackagingSpec(
            artifact_name='test-artifact-name',
            integration_name='test'
        ),
        status=ModelPackagingStatus(
            state=SUCCEEDED_STATE,
        )
    ),
    packaging.packaging,
    'ModelPackaging',
)

PACKAGING_INTEGRATION = EntityTestData(
    PackagingIntegrationClient(),
    PackagingIntegration(
        id=ENTITY_ID,
        spec=PackagingIntegrationSpec(
            default_image="odahu:image",
            entrypoint="some_entrypoint"
        ),
    ),
    packaging_integration.packaging_integration,
    'PackagingIntegration',
)

DEPLOYMENT = EntityTestData(
    ModelDeploymentClient(),
    ModelDeployment(
        id=ENTITY_ID,
        spec=ModelDeploymentSpec(
            image="odahu:image",
            min_replicas=0
        ),
        status=ModelDeploymentStatus(
            state=READY_STATE,
            available_replicas=1,
        )
    ),
    deployment.deployment,
    'ModelDeployment',
)

ROUTER = EntityTestData(
    ModelRouteClient(),
    ModelRoute(
        id=ENTITY_ID,
        spec=ModelRouteSpec(
            mirror="test",
        ),
        status=ModelRouteStatus(
            state=READY_STATE,
        )
    ),
    route.route,
    'ModelRoute',
)



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import http

import click
from odahuflow.cli.utils.client import pass_obj
from odahuflow.cli.utils.error_handler import IGNORE_NOT_FOUND_ERROR_MESSAGE, \
    check_id_or_file_params_present
from odahuflow.cli.utils.output import format_output, DEFAULT_OUTPUT_FORMAT, validate_output_format
from odahuflow.sdk import config
from odahuflow.sdk.clients.api import WrongHttpStatusCode
from odahuflow.sdk.clients.api_aggregated import parse_resources_file_with_one_item
from odahuflow.sdk.clients.toolchain_integration import ToolchainIntegrationClient
from odahuflow.sdk.models import ToolchainIntegration


@click.group()
@click.option('--url', help='API server host', default=config.API_URL)
@click.option('--token', help='API server jwt token', default=config.API_TOKEN)
@click.pass_context
def toolchain_integration(ctx: click.core.Context, url: str, token: str):
    """
    Allow you to perform actions on toolchain integration.\n
    Alias for the command is ti.
    """
    ctx.obj = ToolchainIntegrationClient(url, token)


@toolchain_integration.command()
@click.option('--ti-id', '--id', help='Toolchain integration ID')
@click.option('--output-format', '-o', 'output_format', help='Output format',
              default=DEFAULT_OUTPUT_FORMAT, callback=validate_output_format)
@pass_obj
def get(client: ToolchainIntegrationClient, ti_id: str, output_format: str):
    """
    Get toolchain integrations.\n
    The command without id argument retrieve all toolchain integrations.\n
    Get all toolchain integrations in json format:\n
        odahuflowctl tn-integration get --format json\n
    Get toolchain integration with "git-repo" id:\n
        odahuflowctl tn-integration get --id git-repo\n
    Using jsonpath:\n
        odahuflowctl tn-integration get -o 'jsonpath=[*].spec.reference'
    \f
    :param client: Toolchain integration HTTP client
    :param ti_id: Toolchain integration ID
    :param output_format: Output format
    :return:
    """
    tis = [client.get(ti_id)] if ti_id else client.get_all()

    format_output(tis, output_format)


@toolchain_integration.command()
@click.option('--ti-id', '--id', help='Toolchain integration ID')
@click.option('--file', '-f', type=click.Path(), required=True, help='Path to the file with toolchain integration')
@click.option('--output-format', '-o', 'output_format', help='Output format',
              default=DEFAULT_OUTPUT_FORMAT, callback=validate_output_format)
@pass_obj
def create(client: ToolchainIntegrationClient, ti_id: str, file: str, output_format: str):
    """
    Create a toolchain integration.\n
    You should specify a path to file with a toolchain integration.
    The file must contain only one toolchain integration.
    For now, CLI supports yaml and JSON file formats.
    If you want to create multiples toolchain integrations than you should use "odahuflowctl res apply" instead.
    If you provide the toolchain integration id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl tn-integration create -f ti.yaml --id examples-git
    \f
    :param client: Toolchain integration HTTP client
    :param ti_id: Toolchain integration ID
    :param file: Path to the file with only one toolchain integration
    :param output_format: Output format
    """
    ti = parse_resources_file_with_one_item(file).resource
    if not isinstance(ti, ToolchainIntegration):
        raise ValueError(f'Toolchain integration expected, but {type(ti)} provided')

    if ti_id:
        ti.id = ti_id

    click.echo(format_output([client.create(ti)], output_format))


@toolchain_integration.command()
@click.option('--ti-id', '--id', help='Toolchain integration ID')
@click.option('--file', '-f', type=click.Path(), required=True, help='Path to the file with toolchain integration')
@click.option('--output-format', '-o', 'output_format', help='Output format',
              default=DEFAULT_OUTPUT_FORMAT, callback=validate_output_format)
@pass_obj
def edit(client: ToolchainIntegrationClient, ti_id: str, file: str, output_format: str):
    """
    Update a toolchain integration.\n
    You should specify a path to file with a toolchain integration.
    The file must contain only one toolchain integration.
    For now, CLI supports yaml and JSON file formats.
    If you want to update multiples toolchain integrations than you should use "odahuflowctl res apply" instead.
    If you provide the toolchain integration id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl tn-integration update -f ti.yaml --id examples-git
    \f
    :param client: Toolchain integration HTTP client
    :param ti_id: Toolchain integration ID
    :param file: Path to the file with only one toolchain integration
    :param output_format: Output format
    """
    ti = parse_resources_file_with_one_item(file).resource
    if not isinstance(ti, ToolchainIntegration):
        raise ValueError(f'Toolchain integration expected, but {type(ti)} provided')

    if ti_id:
        ti.id = ti_id

    click.echo(format_output([client.edit(ti)], output_format))


@toolchain_integration.command()
@click.option('--ti-id', '--id', help='Toolchain integration ID')
@click.option('--file', '-f', type=click.Path(), help='Path to the file with toolchain integration')
@click.option('--ignore-not-found/--not-ignore-not-found', default=False,
              help='ignore if toolchain integration is not found')
@pass_obj
def delete(client: ToolchainIntegrationClient, ti_id: str, file: str, ignore_not_found: bool):
    """
    Delete a toolchain integration.\n
    For this command, you must provide a toolchain integration ID or path to file with one toolchain integration.
    The file must contain only one toolchain integration.
    If you want to delete multiples toolchain integrations than you should use "odahuflowctl res delete" instead.
    For now, CLI supports yaml and JSON file formats.
    The command will be failed if you provide both arguments.\n
    Usage example:\n
        * odahuflowctl tn-integration delete --id examples-git\n
        * odahuflowctl tn-integration delete -f ti.yaml
    \f
    :param client: Toolchain integration HTTP client
    :param ti_id: Toolchain integration ID
    :param file: Path to the file with only one toolchain integration
    :param ignore_not_found: ignore if toolchain integration is not found
    """
    check_id_or_file_params_present(ti_id, file)

    if file:
        ti = parse_resources_file_with_one_item(file).resource
        if not isinstance(ti, ToolchainIntegration):
            raise ValueError(f'Toolchain Integration expected, but {type(ti)} provided')

        ti_id = ti.id

    try:
        click.echo(client.delete(ti_id))
    except WrongHttpStatusCode as e:
        if e.status_code != http.HTTPStatus.NOT_FOUND or not ignore_not_found:
            raise e

        click.echo(IGNORE_NOT_FOUND_ERROR_MESSAGE.format(ti_id))



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Training commands for odahuflow cli
"""
import logging
import time
from http.client import HTTPException

import click
from requests import RequestException
from odahuflow.cli.utils.client import pass_obj
from odahuflow.cli.utils.error_handler import check_id_or_file_params_present, TIMEOUT_ERROR_MESSAGE
from odahuflow.cli.utils.logs import print_logs
from odahuflow.cli.utils.output import format_output, DEFAULT_OUTPUT_FORMAT, \
    validate_output_format
from odahuflow.sdk import config
from odahuflow.sdk.clients.api import WrongHttpStatusCode, \
    APIConnectionException
from odahuflow.sdk.clients.api_aggregated import \
    parse_resources_file_with_one_item
from odahuflow.sdk.clients.training import ModelTraining, ModelTrainingClient, \
    TRAINING_SUCCESS_STATE, \
    TRAINING_FAILED_STATE

DEFAULT_WAIT_TIMEOUT = 3
# 1 hour
DEFAULT_TRAINING_TIMEOUT = 60 * 60
LOG_READ_TIMEOUT_SECONDS = 60

LOGGER = logging.getLogger(__name__)


@click.group()
@click.option('--url', help='API server host', default=config.API_URL)
@click.option('--token', help='API server jwt token', default=config.API_TOKEN)
@click.pass_context
def training(ctx: click.core.Context, url: str, token: str):
    """
    Allow you to perform actions on trainings.\n
    Alias for the command is train.
    """
    ctx.obj = ModelTrainingClient(url, token)


@training.command()
@click.option('--train-id', '--id', help='Model training ID')
@click.option('--output-format', '-o', 'output_format', help='Output format',
              default=DEFAULT_OUTPUT_FORMAT, callback=validate_output_format)
@pass_obj
def get(client: ModelTrainingClient, train_id: str, output_format: str):
    """
    Get trainings.\n
    The command without id argument retrieve all trainings.\n
    Get all trainings in json format:\n
        odahuflowctl train get --format json\n
    Get training with "git-repo" id:\n
        odahuflowctl train get --id git-repo\n
    Using jsonpath:\n
        odahuflowctl train get -o 'jsonpath=[*].spec.reference'
    \f
    :param client: Model training HTTP client
    :param train_id: Model training ID
    :param output_format: Output format
    :return:
    """
    trains = [client.get(train_id)] if train_id else client.get_all()

    format_output(trains, output_format)


@training.command()
@click.option('--train-id', '--id', help='Model training ID')
@click.option('--file', '-f', type=click.Path(), required=True,
              help='Path to the file with training')
@click.option('--wait/--no-wait', default=True,
              help='no wait until scale will be finished')
@click.option('--timeout', default=DEFAULT_TRAINING_TIMEOUT, type=int,
              help='timeout in seconds. for wait (if no-wait is off)')
@pass_obj
def create(client: ModelTrainingClient, train_id: str, file: str, wait: bool,
           timeout: int):
    """
    Create a training.\n
    You should specify a path to file with a training. The file must contain only one training.
    For now, CLI supports yaml and JSON file formats.
    If you want to create multiples trainings than you should use "odahuflowctl res apply" instead.
    If you provide the training id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl train create -f train.yaml --id examples-git
    \f
    :param timeout: timeout in seconds. for wait (if no-wait is off)
    :param wait: no wait until scale will be finished
    :param client: Model training HTTP client
    :param train_id: Model training ID
    :param file: Path to the file with only one training
    """
    train = parse_resources_file_with_one_item(file).resource
    if not isinstance(train, ModelTraining):
        raise ValueError(f'ModelTraining expected, but {type(train)} provided')

    if train_id:
        train.id = train_id

    train = client.create(train)
    click.echo(f"Start training: {train}")

    wait_training_finish(timeout, wait, train.id, client)


@training.command()
@click.option('--train-id', '--id', help='Model training ID')
@click.option('--file', '-f', type=click.Path(), required=True,
              help='Path to the file with training')
@click.option('--wait/--no-wait', default=True,
              help='no wait until scale will be finished')
@click.option('--timeout', default=DEFAULT_TRAINING_TIMEOUT, type=int,
              help='timeout in seconds. for wait (if no-wait is off)')
@pass_obj
def edit(client: ModelTrainingClient, train_id: str, file: str, wait: bool,
         timeout: int):
    """
    Rerun a training.\n
    You should specify a path to file with a training. The file must contain only one training.
    For now, CLI supports yaml and JSON file formats.
    If you want to update multiples trainings than you should use "odahuflowctl res apply" instead.
    If you provide the training id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl train update -f train.yaml --id examples-git
    \f
    :param client: Model training HTTP client
    :param train_id: Model training ID
    :param file: Path to the file with only one training
    :param timeout: timeout in seconds. for wait (if no-wait is off)
    :param wait: no wait until scale will be finished
    """
    train = parse_resources_file_with_one_item(file).resource
    if not isinstance(train, ModelTraining):
        raise ValueError(f'Model training expected, but {type(train)} provided')

    if train_id:
        train.id = train_id

    train = client.edit(train)
    click.echo(f"Rerun training: {train}")

    wait_training_finish(timeout, wait, train.id, client)


@training.command()
@click.option('--train-id', '--id', help='Model training ID')
@click.option('--file', '-f', type=click.Path(),
              help='Path to the file with training')
@click.option('--ignore-not-found/--not-ignore-not-found', default=False,
              help='ignore if Model Training is not found')
@pass_obj
def delete(client: ModelTrainingClient, train_id: str, file: str,
           ignore_not_found: bool):
    """
    Delete a training.\n
    For this command, you must provide a training ID or path to file with one training.
    The file must contain only one training.
    If you want to delete multiples trainings than you should use "odahuflowctl res delete" instead.
    For now, CLI supports yaml and JSON file formats.
    The command will be failed if you provide both arguments.\n
    Usage example:\n
        * odahuflowctl train delete --id examples-git\n
        * odahuflowctl train delete -f train.yaml
    \f
    :param client: Model training HTTP client
    :param train_id: Model training ID
    :param file: Path to the file with only one training
    :param ignore_not_found: ignore if Model Training is not found
    """
    check_id_or_file_params_present(train_id, file)

    if file:
        train = parse_resources_file_with_one_item(file).resource
        if not isinstance(train, ModelTraining):
            raise ValueError(
                f'Model training expected, but {type(train)} provided')

        train_id = train.id

    try:
        message = client.delete(train_id)
        click.echo(message)
    except WrongHttpStatusCode as e:
        if e.status_code != 404 or not ignore_not_found:
            raise e

        click.echo(f'Model training {train_id} was not found. Ignore')


@training.command()
@click.option('--train-id', '--id', help='Model training ID')
@click.option('--file', '-f', type=click.Path(),
              help='Path to the file with training')
@click.option('--follow/--not-follow', default=True,
              help='Follow logs stream')
@pass_obj
def logs(client: ModelTrainingClient, train_id: str, file: str, follow: bool):
    """
    Stream training logs.\n
    For this command, you must provide a training ID or path to file with one training.
    The file must contain only one training.
    If you want to delete multiples trainings than you should use "odahuflowctl res delete" instead.
    For now, CLI supports yaml and JSON file formats.
    The command will be failed if you provide both arguments.\n
    Usage example:\n
        * odahuflowctl train delete --id examples-git\n
        * odahuflowctl train delete -f train.yaml
    \f
    :param follow: Follow logs stream
    :param client: Model training HTTP client
    :param train_id: Model training ID
    :param file: Path to the file with only one training
    """
    check_id_or_file_params_present(train_id, file)

    if file:
        train = parse_resources_file_with_one_item(file).resource
        if not isinstance(train, ModelTraining):
            raise ValueError(
                f'Model training expected, but {type(train)} provided')

        train_id = train.id

    for msg in client.log(train_id, follow):
        print_logs(msg)


def wait_training_finish(timeout: int, wait: bool, mt_id: str,
                         mt_client: ModelTrainingClient):
    """
    Wait training to finish according command line arguments

    :param wait:
    :param timeout:
    :param mt_id: Model Training name
    :param mt_client: Model Training Client
    """
    if not wait:
        return

    start = time.time()
    if timeout <= 0:
        raise Exception(
            'Invalid --timeout argument: should be positive integer')

    # We create a separate client for logs because it has the different timeout settings
    log_mt_client = ModelTrainingClient.construct_from_other(mt_client)
    log_mt_client.timeout = mt_client.timeout, LOG_READ_TIMEOUT_SECONDS

    click.echo("Logs streaming...")

    while True:
        elapsed = time.time() - start
        if elapsed > timeout:
            raise Exception(TIMEOUT_ERROR_MESSAGE)

        try:
            mt = mt_client.get(mt_id)
            if mt.status.state == TRAINING_SUCCESS_STATE:
                click.echo(
                    f'Model {mt_id} was trained. Training took {round(time.time() - start)} seconds')
                return
            elif mt.status.state == TRAINING_FAILED_STATE:
                raise Exception(f'Model training {mt_id} was failed.')
            elif mt.status.state == "":
                click.echo(f"Can't determine the state of {mt.id}. Sleeping...")
            else:
                for msg in log_mt_client.log(mt.id, follow=True):
                    print_logs(msg)

        except (WrongHttpStatusCode, HTTPException, RequestException,
                APIConnectionException) as e:
            LOGGER.info(
                'Callback have not confirmed completion of the operation. Exception: %s',
                str(e))

        LOGGER.debug('Sleep before next request')
        time.sleep(DEFAULT_WAIT_TIMEOUT)



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Abbreviation Group for click
"""
import click

ABBREVIATION = {
    "conn": "connection",
    "conf": "config",
    "dep": "deployment",
    "pack": "packaging",
    "pi": "packaging-integration",
    "temp": "template",
    "ti": "toolchain-integration",
    "train": "training"
}


class AbbreviationGroup(click.Group):
    """
    AbbreviationGroup
    """

    def get_command(self, ctx, cmd_name):
        """
        Override get command of click.Group
        :param ctx: click context
        :param cmd_name: group name
        :return: click Command
        """
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        return click.Group.get_command(self, ctx, ABBREVIATION.get(cmd_name))



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import http

import click
from odahuflow.cli.utils.client import pass_obj
from odahuflow.cli.utils.error_handler import check_id_or_file_params_present
from odahuflow.cli.utils.output import format_output, DEFAULT_OUTPUT_FORMAT, validate_output_format
from odahuflow.sdk import config
from odahuflow.sdk.clients.api import WrongHttpStatusCode
from odahuflow.sdk.clients.api_aggregated import parse_resources_file_with_one_item
from odahuflow.sdk.clients.packaging_integration import PackagingIntegrationClient
from odahuflow.sdk.models import PackagingIntegration

IGNORE_NOT_FOUND_ERROR_MESSAGE = 'Packaging integration {} was not found. Ignore'
ID_AND_FILE_MISSED_ERROR_MESSAGE = f'You should provide a packaging integration ID or file parameter, not both.'


@click.group()
@click.option('--url', help='API server host', default=config.API_URL)
@click.option('--token', help='API server jwt token', default=config.API_TOKEN)
@click.pass_context
def packaging_integration(ctx: click.core.Context, url: str, token: str):
    """
    Allow you to perform actions on packaging integration.\n
    Alias for the command is pi.
    """
    ctx.obj = PackagingIntegrationClient(url, token)


@packaging_integration.command()
@click.option('--pi-id', '--id', help='Packaging integration ID')
@click.option('--output-format', '-o', 'output_format', help='Output format',
              default=DEFAULT_OUTPUT_FORMAT, callback=validate_output_format)
@pass_obj
def get(client: PackagingIntegrationClient, pi_id: str, output_format: str):
    """
    Get packaging integrations.\n
    The command without id argument retrieve all packaging integrations.\n
    Get all packaging integrations in json format:\n
        odahuflowctl pack-integration get --format json\n
    Get packaging integration with "git-repo" id:\n
        odahuflowctl pack-integration get --id git-repo\n
    Using jsonpath:\n
        odahuflowctl pack-integration get -o 'jsonpath=[*].spec.reference'
    \f
    :param client: Packaging integration HTTP client
    :param pi_id: Packaging integration ID
    :param output_format: Output format
    :return:
    """
    pis = [client.get(pi_id)] if pi_id else client.get_all()

    format_output(pis, output_format)


@packaging_integration.command()
@click.option('--pi-id', '--id', help='Packaging integration ID')
@click.option('--file', '-f', type=click.Path(), required=True, help='Path to the file with packaging integration')
@click.option('--output-format', '-o', 'output_format', help='Output format',
              default=DEFAULT_OUTPUT_FORMAT, callback=validate_output_format)
@pass_obj
def create(client: PackagingIntegrationClient, pi_id: str, file: str, output_format: str):
    """
    Create a packaging integration.\n
    You should specify a path to file with a packaging integration.
    The file must contain only one packaging integration.
    For now, CLI supports yaml and JSON file formats.
    If you want to create multiples packaging integrations than you should use "odahuflowctl res apply" instead.
    If you provide the packaging integration id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl pack-integration create -f pi.yaml --id examples-git
    \f
    :param client: Packaging integration HTTP client
    :param pi_id: Packaging integration ID
    :param file: Path to the file with only one packaging integration
    :param output_format: Output format
    """
    pi = parse_resources_file_with_one_item(file).resource
    if not isinstance(pi, PackagingIntegration):
        raise ValueError(f'Packaging integration expected, but {type(pi)} provided')

    if pi_id:
        pi.id = pi_id

    click.echo(format_output([client.create(pi)], output_format))


@packaging_integration.command()
@click.option('--pi-id', '--id', help='Packaging integration ID')
@click.option('--file', '-f', type=click.Path(), required=True, help='Path to the file with packaging integration')
@click.option('--output-format', '-o', 'output_format', help='Output format',
              default=DEFAULT_OUTPUT_FORMAT, callback=validate_output_format)
@pass_obj
def edit(client: PackagingIntegrationClient, pi_id: str, file: str, output_format: str):
    """
    Update a packaging integration.\n
    You should specify a path to file with a packaging integration.
    The file must contain only one packaging integration.
    For now, CLI supports yaml and JSON file formats.
    If you want to update multiples packaging integrations than you should use "odahuflowctl res apply" instead.
    If you provide the packaging integration id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl pack-integration update -f pi.yaml --id examples-git
    \f
    :param client: Packaging integration HTTP client
    :param pi_id: Packaging integration ID
    :param file: Path to the file with only one packaging integration
    :param output_format: Output format
    """
    pi = parse_resources_file_with_one_item(file).resource
    if not isinstance(pi, PackagingIntegration):
        raise ValueError(f'Packaging integration expected, but {type(pi)} provided')

    if pi_id:
        pi.id = pi_id

    click.echo(format_output([client.edit(pi)], output_format))


@packaging_integration.command()
@click.option('--pi-id', '--id', help='Packaging integration ID')
@click.option('--file', '-f', type=click.Path(), help='Path to the file with packaging integration')
@click.option('--ignore-not-found/--not-ignore-not-found', default=False,
              help='ignore if toolchain integration is not found')
@pass_obj
def delete(client: PackagingIntegrationClient, pi_id: str, file: str, ignore_not_found: bool):
    """
    Delete a packaging integration.\n
    For this command, you must provide a packaging integration ID or path to file with one packaging integration.
    The file must contain only one packaging integration.
    If you want to delete multiples packaging integrations than you should use "odahuflowctl res delete" instead.
    For now, CLI supports yaml and JSON file formats.
    The command will be failed if you provide both arguments.\n
    Usage example:\n
        * odahuflowctl pack-integration delete --id examples-git\n
        * odahuflowctl pack-integration delete -f pi.yaml
    \f
    :param client: Packaging integration HTTP client
    :param pi_id: Packaging integration ID
    :param file: Path to the file with only one packaging integration
    :param ignore_not_found: ignore if toolchain integration is not found
    """
    check_id_or_file_params_present(pi_id, file)

    if file:
        pi = parse_resources_file_with_one_item(file).resource
        if not isinstance(pi, PackagingIntegration):
            raise ValueError(f'Packaging Integration expected, but {type(pi)} provided')

        pi_id = pi.id

    try:
        click.echo(client.delete(pi_id))
    except WrongHttpStatusCode as e:
        if e.status_code != http.HTTPStatus.NOT_FOUND or not ignore_not_found:
            raise e

        click.echo(IGNORE_NOT_FOUND_ERROR_MESSAGE.format(pi_id))



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
API commands for odahuflow cli
"""
import logging
import time

import click
from odahuflow.cli.utils.client import pass_obj
from odahuflow.cli.utils.error_handler import check_id_or_file_params_present, TIMEOUT_ERROR_MESSAGE
from odahuflow.cli.utils.output import DEFAULT_OUTPUT_FORMAT, format_output, validate_output_format
from odahuflow.cli.utils.verifiers import positive_number
from odahuflow.sdk import config
from odahuflow.sdk.clients.deployment import ModelDeployment, ModelDeploymentClient, READY_STATE, \
    FAILED_STATE
from odahuflow.sdk.clients.api import WrongHttpStatusCode
from odahuflow.sdk.clients.api_aggregated import parse_resources_file_with_one_item

DEFAULT_WAIT_TIMEOUT = 5
# 20 minutes
DEFAULT_DEPLOYMENT_TIMEOUT = 20 * 60

DEFAULT_WIDTH = 120
MD_HEADER = ["Name", "State", "Min/Current/Max Replicas", "Service URL"]

DEFAULT_WIDTH_LOCAL = 160
DELETION_TIMEOUT = 60
MD_LOCAL_HEADER = ["Name", "Image", "Port", "Model Name", "Model Version"]

LOGGER = logging.getLogger(__name__)


@click.group()
@click.option('--url', help='API server host', default=config.API_URL)
@click.option('--token', help='API server jwt token', default=config.API_TOKEN)
@click.pass_context
def deployment(ctx: click.core.Context, url: str, token: str):
    """
    Allow you to perform actions on deployments.\n
    Alias for the command is dep.
    """
    ctx.obj = ModelDeploymentClient(url, token)


@deployment.command()
@click.option('--md-id', '--id', help='Model deployment ID')
@click.option('--output-format', '-o', 'output_format', help='Output format',
              default=DEFAULT_OUTPUT_FORMAT, callback=validate_output_format)
@pass_obj
def get(client: ModelDeploymentClient, md_id: str, output_format: str):
    """
    Get deployments.\n
    The command without id argument retrieve all deployments.\n
    Get all deployments in json format:\n
        odahuflowctl dep get --format json\n
    Get deployment with "git-repo" id:\n
        odahuflowctl dep get --id model-wine\n
    Using jsonpath:\n
        odahuflowctl dep get -o 'jsonpath=[*].spec.reference'
    \f
    :param client: Model deployment HTTP client
    :param md_id: Model deployment ID
    :param output_format: Output format
    :return:
    """
    mds = [client.get(md_id)] if md_id else client.get_all()

    format_output(mds, output_format)


@deployment.command()
@click.option('--md-id', '--id', help='Model deployment ID')
@click.option('--file', '-f', type=click.Path(), required=True, help='Path to the file with deployment')
@click.option('--wait/--no-wait', default=True,
              help='no wait until scale will be finished')
@click.option('--timeout', default=DEFAULT_DEPLOYMENT_TIMEOUT, type=int, callback=positive_number,
              help='timeout in seconds. for wait (if no-wait is off)')
@click.option('--image', type=str, help='Override Docker image from file')
@pass_obj
def create(client: ModelDeploymentClient, md_id: str, file: str, wait: bool, timeout: int, image: str):
    """
    Create a deployment.\n
    You should specify a path to file with a deployment. The file must contain only one deployment.
    For now, CLI supports yaml and JSON file formats.
    If you want to create multiples deployments than you should use "odahuflowctl res apply" instead.
    If you provide the deployment id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl dep create -f dep.yaml --id examples-git
    \f
    :param timeout: timeout in seconds. for wait (if no-wait is off)
    :param wait: no wait until deployment will be finished
    :param client: Model deployment HTTP client
    :param md_id: Model deployment ID
    :param file: Path to the file with only one deployment
    :param image: Override Docker image from file
    """
    md = parse_resources_file_with_one_item(file).resource
    if not isinstance(md, ModelDeployment):
        raise ValueError(f'Model deployment expected, but {type(md)} provided')

    if md_id:
        md.id = md_id

    if image:
        md.spec.image = image

    click.echo(client.create(md))

    wait_deployment_finish(timeout, wait, md.id, client)


@deployment.command()
@click.option('--md-id', '--id', help='Model deployment ID')
@click.option('--file', '-f', type=click.Path(), required=True, help='Path to the file with deployment')
@click.option('--wait/--no-wait', default=True,
              help='no wait until scale will be finished')
@click.option('--timeout', default=DEFAULT_DEPLOYMENT_TIMEOUT, type=int, callback=positive_number,
              help='timeout in seconds. for wait (if no-wait is off)')
@click.option('--image', type=str, help='Override Docker image from file')
@pass_obj
def edit(client: ModelDeploymentClient, md_id: str, file: str, wait: bool, timeout: int, image: str):
    """
    Update a deployment.\n
    You should specify a path to file with a deployment. The file must contain only one deployment.
    For now, CLI supports yaml and JSON file formats.
    If you want to update multiples deployments than you should use "odahuflowctl res apply" instead.
    If you provide the deployment id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl dep update -f dep.yaml --id examples-git
    \f
    :param client: Model deployment HTTP client
    :param md_id: Model deployment ID
    :param file: Path to the file with only one deployment
    :param timeout: timeout in seconds. for wait (if no-wait is off)
    :param wait: no wait until edit will be finished
    :param image: Override Docker image from file
    """
    md = parse_resources_file_with_one_item(file).resource
    if not isinstance(md, ModelDeployment):
        raise ValueError(f'Model deployment expected, but {type(md)} provided')

    if md_id:
        md.id = md_id

    if image:
        md.spec.image = image

    click.echo(client.edit(md))

    wait_deployment_finish(timeout, wait, md.id, client)


@deployment.command()
@click.option('--md-id', '--id', help='Model deployment ID')
@click.option('--file', '-f', type=click.Path(), help='Path to the file with deployment')
@click.option('--wait/--no-wait', default=True,
              help='no wait until scale will be finished')
@click.option('--timeout', default=DEFAULT_DEPLOYMENT_TIMEOUT, type=int, callback=positive_number,
              help='timeout in seconds. for wait (if no-wait is off)')
@click.option('--ignore-not-found/--not-ignore-not-found', default=False,
              help='ignore if Model Deployment is not found')
@pass_obj
def delete(client: ModelDeploymentClient, md_id: str, file: str, ignore_not_found: bool,
           wait: bool, timeout: int):
    """
    Delete a deployment.\n
    For this command, you must provide a deployment ID or path to file with one deployment.
    The file must contain only one deployment.
    If you want to delete multiples deployments than you should use "odahuflowctl res delete" instead.
    For now, CLI supports yaml and JSON file formats.
    The command will be failed if you provide both arguments.\n
    Usage example:\n
        * odahuflowctl dep delete --id examples-git\n
        * odahuflowctl dep delete -f dep.yaml
    \f
    :param timeout: timeout in seconds. for wait (if no-wait is off)
    :param wait: no wait until deletion will be finished
    :param client: Model deployment HTTP client
    :param md_id: Model deployment ID
    :param file: Path to the file with only one deployment
    :param ignore_not_found: ignore if Model Deployment is not found
    """
    check_id_or_file_params_present(md_id, file)

    if file:
        md = parse_resources_file_with_one_item(file).resource
        if not isinstance(md, ModelDeployment):
            raise ValueError(f'Model deployment expected, but {type(md)} provided')

        md_id = md.id

    try:
        message = client.delete(md_id)

        wait_delete_operation_finish(timeout, wait, md_id, client)
        click.echo(message)
    except WrongHttpStatusCode as e:
        if e.status_code != 404 or not ignore_not_found:
            raise e

        click.echo(f'Model deployment {md_id} was not found. Ignore')


def wait_delete_operation_finish(timeout: int, wait: bool, md_id: str, md_client: ModelDeploymentClient):
    """
    Wait delete operation

    :param timeout: timeout in seconds. for wait (if no-wait is off)
    :param wait: no wait until deletion will be finished
    :param md_id: Model Deployment name
    :param md_client: Model Deployment Client

    :return: None
    """
    if not wait:
        return

    start = time.time()
    if timeout <= 0:
        raise Exception('Invalid --timeout argument: should be positive integer')

    while True:
        elapsed = time.time() - start
        if elapsed > timeout:
            raise Exception('Time out: operation has not been confirmed')

        try:
            md_client.get(md_id)
        except WrongHttpStatusCode as e:
            if e.status_code == 404:
                print(f'Model deployment {md_id} was deleted')
                return
            LOGGER.info('Callback have not confirmed completion of the operation')

        print(f'Model deployment {md_id} is still being deleted...')
        time.sleep(DEFAULT_WAIT_TIMEOUT)


def wait_deployment_finish(timeout: int, wait: bool, md_id: str, md_client: ModelDeploymentClient):
    """
    Wait deployment to finish according command line arguments

    :param timeout: timeout in seconds. for wait (if no-wait is off)
    :param wait: no wait until deletion will be finished
    :param md_id: Model Deployment name
    :param md_client: Model Deployment Client

    :return: None
    """
    if not wait:
        return

    start = time.time()
    if timeout <= 0:
        raise Exception('Invalid --timeout argument: should be positive integer')

    while True:
        elapsed = time.time() - start
        if elapsed > timeout:
            raise Exception(TIMEOUT_ERROR_MESSAGE)

        try:
            md: ModelDeployment = md_client.get(md_id)
            if md.status.state == READY_STATE:
                if md.spec.min_replicas <= md.status.available_replicas:
                    print(f'Model {md_id} was deployed. '
                          f'Deployment process took is {round(time.time() - start)} seconds')
                    return
                else:
                    print(f'Model {md_id} was deployed. '
                          f'Number of available pods is {md.status.available_replicas}/{md.spec.min_replicas}')
            elif md.status.state == FAILED_STATE:
                raise Exception(f'Model deployment {md_id} was failed')
            elif md.status.state == "":
                print(f"Can't determine the state of {md.id}. Sleeping...")
            else:
                print(f'Current deployment state is {md.status.state}. Sleeping...')
        except WrongHttpStatusCode:
            LOGGER.info('Callback have not confirmed completion of the operation')

        LOGGER.debug('Sleep before next request')
        time.sleep(DEFAULT_WAIT_TIMEOUT)



#
#    Copyright 2020 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import click
import click_completion
import click_completion.core


@click.command()
@click.argument('shell', required=False, type=click_completion.DocumentedChoice(
    click_completion.core.shells))
def completion(shell):
    """
    Output odahuflowctl completion code to stdout.\n
    Load the zsh completion in the current shell:\n
        source <(odahuflowctl completion zsh)
    \f
    """
    click.echo(click_completion.core.get_code(shell))



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
CLI entrypoint
"""

import click
import click_completion
from odahuflow.cli.parsers.bulk import bulk
from odahuflow.cli.parsers.completion import completion
from odahuflow.cli.parsers.config import config_group
from odahuflow.cli.parsers.connection import connection
from odahuflow.cli.parsers.deployment import deployment
from odahuflow.cli.parsers.gppi import gppi
from odahuflow.cli.parsers.local import local
from odahuflow.cli.parsers.model import model
from odahuflow.cli.parsers.packaging import packaging
from odahuflow.cli.parsers.packaging_integration import packaging_integration
from odahuflow.cli.parsers.route import route
from odahuflow.cli.parsers.security import login, logout
from odahuflow.cli.parsers.template import template
from odahuflow.cli.parsers.toolchain_integration import toolchain_integration
from odahuflow.cli.parsers.training import training
from odahuflow.cli.utils.abbr import AbbreviationGroup
from odahuflow.cli.utils.error_handler import cli_error_handler
from odahuflow.sdk.logger import configure_logging
from odahuflow.cli.version import version

COMMAND_GROUPS = [
    config_group,
    connection,
    deployment,
    model,
    packaging,
    packaging_integration,
    bulk,
    route,
    template,
    gppi,
    toolchain_integration,
    training,
    login,
    logout,
    version,
    completion,
    local,
]

# Initialize shell completion
click_completion.init()


@click.group(cls=AbbreviationGroup)
@click.option('--verbose/--non-verbose', default=False)
def base(verbose=False):
    """
    Odahuflow CLI
    """
    configure_logging(verbose)


# Initialize all commands
for group in COMMAND_GROUPS:
    base.add_command(group)


def main():
    """
    Main CLI entrypoint
    """
    with cli_error_handler():
        base()


if __name__ == '__main__':
    base()



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
API commands for odahuflow cli
"""
import logging
import time

import click
from click import pass_obj
from odahuflow.cli.utils.error_handler import check_id_or_file_params_present, TIMEOUT_ERROR_MESSAGE
from odahuflow.cli.utils.output import DEFAULT_OUTPUT_FORMAT, format_output, validate_output_format
from odahuflow.sdk import config
from odahuflow.sdk.clients.api import WrongHttpStatusCode
from odahuflow.sdk.clients.api_aggregated import parse_resources_file_with_one_item
from odahuflow.sdk.clients.route import ModelRoute, ModelRouteClient, READY_STATE

DEFAULT_WAIT_TIMEOUT = 5
LOGGER = logging.getLogger(__name__)


@click.group()
@click.option('--url', help='API server host', default=config.API_URL)
@click.option('--token', help='API server jwt token', default=config.API_TOKEN)
@click.pass_context
def route(ctx: click.core.Context, url: str, token: str):
    """
    Allow you to perform actions on routes
    """
    ctx.obj = ModelRouteClient(url, token)


@route.command()
@click.option('--mr-id', '--id', 'mr_id', help='ModelRoute ID')
@click.option('--output-format', '-o', 'output_format', help='Output format',
              default=DEFAULT_OUTPUT_FORMAT, callback=validate_output_format)
@pass_obj
def get(client: ModelRouteClient, mr_id: str, output_format: str):
    """
    Get routes.\n
    The command without id argument retrieve all routes.\n
    Get all routes in json format:\n
        odahuflowctl conn get --format json\n
    Get model route with "git-repo" id:\n
        odahuflowctl conn get --id git-repo\n
    Using jsonpath:\n
        odahuflowctl conn get -o 'jsonpath=[*].spec.reference'
    \f
    :param client: ModelRoute HTTP client
    :param mr_id: ModelRoute ID
    :param output_format: Output format
    :return:
    """
    conns = [client.get(mr_id)] if mr_id else client.get_all()

    format_output(conns, output_format)


@route.command()
@click.option('--mr-id', '--id', 'mr_id', help='ModelRoute ID')
@click.option('--file', '-f', type=click.Path(), required=True, help='Path to the file with model route')
@click.option('--wait/--no-wait', default=True,
              help='no wait until scale will be finished')
@click.option('--timeout', default=600, type=int,
              help='timeout in seconds. for wait (if no-wait is off)')
@pass_obj
def create(client: ModelRouteClient, mr_id: str, file: str, wait: bool, timeout: int):
    """
    Create a model route.\n
    You should specify a path to file with a model route. The file must contain only one model route.
    For now, CLI supports yaml and JSON file formats.
    If you want to create multiples routes than you should use "odahuflowctl res apply" instead.
    If you provide the model route id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl conn create -f conn.yaml --id examples-git
    \f
    :param timeout: timeout in seconds. for wait (if no-wait is off)
    :param wait: no wait until operation will be finished
    :param client: ModelRoute HTTP client
    :param mr_id: ModelRoute ID
    :param file: Path to the file with only one model route
    """
    conn = parse_resources_file_with_one_item(file).resource
    if not isinstance(conn, ModelRoute):
        raise ValueError(f'ModelRoute expected, but {type(conn)} provided')

    if mr_id:
        conn.id = mr_id

    click.echo(client.create(conn))

    wait_operation_finish(timeout, wait, mr_id, client)


@route.command()
@click.option('--route-id', '--id', 'mr_id', help='ModelRoute ID')
@click.option('--file', '-f', type=click.Path(), required=True, help='Path to the file with model route')
@click.option('--wait/--no-wait', default=True,
              help='no wait until scale will be finished')
@click.option('--timeout', default=600, type=int,
              help='timeout in seconds. for wait (if no-wait is off)')
@pass_obj
def edit(client: ModelRouteClient, mr_id: str, file: str, wait: bool, timeout: int):
    """
    Update a model route.\n
    You should specify a path to file with a model route. The file must contain only one model route.
    For now, CLI supports yaml and JSON file formats.
    If you want to update multiples routes than you should use "odahuflowctl res apply" instead.
    If you provide the model route id parameter than it will be overridden before sending to API server.\n
    Usage example:\n
        * odahuflowctl conn update -f conn.yaml --id examples-git
    \f
    :param timeout: timeout in seconds. for wait (if no-wait is off)
    :param wait: no wait until operation will be finished
    :param client: Model route HTTP client
    :param mr_id: Model route ID
    :param file: Path to the file with only one model route
    """
    conn = parse_resources_file_with_one_item(file).resource
    if not isinstance(conn, ModelRoute):
        raise ValueError(f'ModelRoute expected, but {type(conn)} provided')

    if mr_id:
        conn.id = mr_id

    click.echo(client.edit(conn))

    wait_operation_finish(timeout, wait, mr_id, client)


@route.command()
@click.option('--route-id', '--id', 'mr_id', help='ModelRoute ID')
@click.option('--file', '-f', type=click.Path(), help='Path to the file with model route')
@click.option('--ignore-not-found/--not-ignore-not-found', default=False,
              help='ignore if Model Deployment is not found')
@pass_obj
def delete(client: ModelRouteClient, mr_id: str, file: str, ignore_not_found: bool):
    """
    Delete a model route.\n
    For this command, you must provide a model route ID or path to file with one model route.
    The file must contain only one model route.
    If you want to delete multiples routes than you should use "odahuflowctl res delete" instead.
    For now, CLI supports yaml and JSON file formats.
    The command will be failed if you provide both arguments.\n
    Usage example:\n
        * odahuflowctl conn delete --id examples-git\n
        * odahuflowctl conn delete -f conn.yaml
    \f
    :param client: ModelRoute HTTP client
    :param mr_id: ModelRoute ID
    :param file: Path to the file with only one model route
    :param ignore_not_found: ignore if Model Deployment is not found
    """
    check_id_or_file_params_present(mr_id, file)

    if file:
        conn = parse_resources_file_with_one_item(file).resource
        if not isinstance(conn, ModelRoute):
            raise ValueError(f'ModelRoute expected, but {type(conn)} provided')

        mr_id = conn.id

    try:
        click.echo(client.delete(mr_id))
    except WrongHttpStatusCode as e:
        if e.status_code != 404 or not ignore_not_found:
            raise e

        click.echo(f'Model route {mr_id} was not found. Ignore')


def wait_operation_finish(timeout: int, wait: bool, mr_id: str, mr_client: ModelRouteClient):
    """
    Wait route to finish according command line arguments

    :param timeout: timeout in seconds. for wait (if no-wait is off)
    :param wait: no wait until operation will be finished
    :param mr_id: Model Route id
    :param mr_client: Model Route Client

    :return: None
    """
    if not wait:
        return

    start = time.time()
    if timeout <= 0:
        raise Exception('Invalid --timeout argument: should be positive integer')

    while True:
        elapsed = time.time() - start
        if elapsed > timeout:
            raise Exception(TIMEOUT_ERROR_MESSAGE)

        try:
            mr = mr_client.get(mr_id)
            if mr.status.state == READY_STATE:
                print(f'Model Route {mr_id} is ready')
                return
            elif mr.status.state == "":
                print(f"Can't determine the state of {mr.id}. Sleeping...")
            else:
                print(f'Current route state is {mr.status.state}. Sleeping...')
        except WrongHttpStatusCode:
            LOGGER.info('Callback have not confirmed completion of the operation')

        LOGGER.debug('Sleep before next request')
        time.sleep(DEFAULT_WAIT_TIMEOUT)



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Security commands for odahuflow cli
"""
import logging
import sys

import click
from click import UsageError

import odahuflow.sdk.config as config
from odahuflow.sdk.clients import api
from odahuflow.sdk.clients.oidc import OpenIdProviderConfiguration
from odahuflow.sdk.config import update_config_file

LOG = logging.getLogger(__name__)


def _reset_credentials():
    """
    Clean credentials from config file
    :return:
    """
    update_config_file(API_URL=None,
                       API_TOKEN=None,
                       API_REFRESH_TOKEN=None,
                       API_ACCESS_TOKEN=None,
                       API_ISSUING_URL=None,
                       ODAHUFLOWCTL_OAUTH_CLIENT_ID=None,
                       ODAHUFLOWCTL_OAUTH_CLIENT_SECRET=None,
                       ISSUER_URL=None)


def fetch_openid_conf(issuer: str) -> None:
    conf = OpenIdProviderConfiguration(issuer)
    conf.fetch_configuration()

    config.API_ISSUING_URL = conf.token_endpoint


@click.command()
@click.option('--url', 'api_host', help='API server host', required=True)
@click.option('--token', help='API server jwt token')
@click.option('--client_id', help='client_id for OAuth2 Client Credentials flow')
@click.option('--client_secret', help='client_secret for OAuth2 Client Credentials flow')
@click.option('--issuer', help='OIDC Issuer URL')
def login(api_host: str, token: str, client_id: str, client_secret: str, issuer: str):
    """
    Authorize on API endpoint.
    Check that credentials is correct and save to the config
    """

    # clean config from previous credentials
    _reset_credentials()

    # update config
    update_config_file(
        API_URL=api_host,
        ODAHUFLOWCTL_OAUTH_CLIENT_ID=client_id,
        ODAHUFLOWCTL_OAUTH_CLIENT_SECRET=client_secret,
        API_TOKEN=token,
        ISSUER_URL=issuer
    )

    # set predicates
    is_token_login = bool(token)
    is_client_cred_login = bool(client_id) or bool(client_secret)
    is_interactive_login = not (is_token_login or is_client_cred_login)

    # validate
    if is_token_login and is_client_cred_login:
        raise UsageError('You should use either --token or --client_id/--client_secret to login. '
                         'Otherwise skipp all options to launch interactive login mode')
    if is_client_cred_login and (not client_id or not client_secret):
        raise UsageError('You must pass both client_id and client_secret to client_credentials login')
    if is_client_cred_login and not config.ISSUER_URL:
        raise UsageError('You must provide --issuer parameter to do client_credentials login. '
                         'Or set ISSUER_URL option in config file')

    # fetch openid configuration
    if config.ISSUER_URL:
        fetch_openid_conf(config.ISSUER_URL)

    # login
    try:
        api_client = api.RemoteAPIClient(api_host, token, client_id, client_secret,
                                         non_interactive=not is_interactive_login)

        api_client.info()

        print('Success! Credentials have been saved.')
    except api.IncorrectAuthorizationToken as wrong_token:
        LOG.error('Wrong authorization token\n%s', wrong_token)
        sys.exit(1)


@click.command()
def logout():
    """
    Remove all authorization data from the configuration file
    """
    _reset_credentials()
    click.echo('All authorization credentials have been removed')
